import os
import re
from typing import Literal

import httpx
from langchain_google_community import GoogleSearchAPIWrapper
from raggy.vectorstores.chroma import query_collection
from settings import settings


def convert_md_links_to_slack(text) -> str:
    md_link_pattern = r"\[(?P<text>[^\]]+)]\((?P<url>[^\)]+)\)"

    def to_slack_link(match):
        return f'<{match.group("url")}|{match.group("text")}>'

    return re.sub(
        r"\*\*(.*?)\*\*", r"*\1*", re.sub(md_link_pattern, to_slack_link, text)
    )


async def post_slack_message(
    message: str,
    channel_id: str,
    attachments: list | None = None,
    thread_ts: str | None = None,
    auth_token: str | None = None,
) -> dict:
    async with httpx.AsyncClient() as client:
        response = await client.post(
            "https://slack.com/api/chat.postMessage",
            headers={
                "Authorization": f"Bearer {auth_token or os.environ["SLACK_API_TOKEN"]}"
            },
            json={
                "channel": channel_id,
                "text": convert_md_links_to_slack(message),
                "attachments": attachments if attachments else [],
                **({"thread_ts": thread_ts} if thread_ts else {}),
            },
        )
    data = response.json()
    if data.get("ok") is not True:
        raise ValueError(f"Error posting Slack message: {data.get('error')}")
    return data


## tools
def search_internet(query: str) -> str:
    """Search the internet for information relevant to the query."""
    return GoogleSearchAPIWrapper(
        google_api_key=settings.google_api_key.get_secret_value(),
        google_cse_id=settings.google_cse_id.get_secret_value(),
    ).run(query)


async def search_knowledge_base(query: str, domain: Literal["docs"]) -> str:
    """Search documentation for information relevant to the query."""
    return await query_collection(
        query_text=query,
        collection_name=domain,
        client_type=settings.chroma_client_type,
    )



import controlflow as cf

classifier = cf.Agent(model="openai/gpt-4o-mini")


def classify_news(headline: str) -> str:
    return cf.run(
        "Classify the news headline into the most appropriate category",
        agents=[classifier],
        result_type=["Politics", "Technology", "Sports", "Entertainment", "Science"],
        context={"headline": headline},
    )


if __name__ == "__main__":
    headline = "New AI Model Breaks Records in Language Understanding"
    category = classify_news(headline)
    print(f"Headline: {headline}")
    print(f"Category: {category}")

    headline = "Scientists Discover Potentially Habitable Exoplanet"
    category = classify_news(headline)
    print(f"\nHeadline: {headline}")
    print(f"Category: {category}")



import re
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, Optional

import tiktoken

import controlflow
from controlflow.events.base import Event, UnpersistedEvent
from controlflow.events.events import (
    AgentMessage,
    ToolCallEvent,
    ToolResultEvent,
)
from controlflow.llm.messages import (
    AIMessage,
    BaseMessage,
    HumanMessage,
    SystemMessage,
)
from controlflow.llm.rules import LLMRules
from controlflow.utilities.logging import get_logger

if TYPE_CHECKING:
    from controlflow.agents.agent import Agent
logger = get_logger(__name__)


class CombinedAgentMessage(UnpersistedEvent):
    event: Literal["combined-agent-message"] = "combined-agent-message"
    agent_message: AgentMessage
    tool_call: list[ToolCallEvent] = []
    tool_results: list[ToolResultEvent] = []

    def to_messages(self, context: "CompileContext") -> list[BaseMessage]:
        messages = []
        messages.extend(self.agent_message.to_messages(context))
        for tool_result in self.tool_results:
            messages.extend(tool_result.to_messages(context))
        return messages


def add_user_message_to_beginning(
    messages: list[BaseMessage], rules: LLMRules
) -> list[BaseMessage]:
    """
    If the LLM requires the user message to be the first message, add a user
    message to the beginning of the list.
    """
    if rules.require_user_message_after_system:
        if not messages or not isinstance(messages[0], HumanMessage):
            messages.insert(0, HumanMessage(content="SYSTEM: Begin."))
    return messages


def ensure_at_least_one_message(
    messages: list[BaseMessage], rules: LLMRules
) -> list[BaseMessage]:
    if not messages and rules.require_at_least_one_message:
        messages.append(HumanMessage(content="SYSTEM: Begin."))
    return messages


def add_user_message_to_end(
    messages: list[BaseMessage], rules: LLMRules
) -> list[BaseMessage]:
    """
    If the LLM doesn't allow the last message to be from the AI when using tools,
    add a user message to the end of the list.
    """
    if not rules.allow_last_message_from_ai_when_using_tools:
        if not messages or isinstance(messages[-1], AIMessage):
            msg = HumanMessage(content="SYSTEM: Continue.")
            messages.append(msg)
    return messages


def remove_duplicate_messages(messages: list[BaseMessage]) -> list[BaseMessage]:
    """
    Removes duplicate messages from the list.
    """
    seen = set()
    new_messages = []
    for message in messages:
        if message.id not in seen:
            new_messages.append(message)
            if message.id:
                seen.add(message.id)
    return new_messages


def break_up_consecutive_ai_messages(
    messages: list[BaseMessage], rules: LLMRules
) -> list[BaseMessage]:
    """
    Breaks up consecutive AI messages by inserting a system message.
    """
    if not messages or rules.allow_consecutive_ai_messages:
        return messages

    new_messages = messages.copy()
    i = 1
    while i < len(new_messages):
        if isinstance(new_messages[i], AIMessage) and isinstance(
            new_messages[i - 1], AIMessage
        ):
            new_messages.insert(i, SystemMessage(content="Continue."))
        i += 1

    return new_messages


def convert_system_messages(
    messages: list[BaseMessage], rules: LLMRules
) -> list[BaseMessage]:
    """
    Converts system messages to human messages if the LLM doesnt support system
    messages, either at all or in the first position.
    """
    new_messages = []
    for i, message in enumerate(messages):
        if isinstance(message, SystemMessage):
            # If system messages are not supported OR if they must be first and
            # this is not the first message, THEN convert the message to a human message
            if not rules.allow_system_messages or (
                i > 0 and rules.require_system_message_first
            ):
                new_messages.append(
                    HumanMessage(
                        content=f"ORCHESTRATOR: {message.content}", name=message.name
                    )
                )
            else:
                # If the system message is allowed, add it as-is
                new_messages.append(message)
        else:
            new_messages.append(message)
    return new_messages


def format_message_name(
    messages: list[BaseMessage], rules: LLMRules
) -> list[BaseMessage]:
    if not rules.require_message_name_format:
        return messages

    for message in messages:
        if message.name:
            name = re.sub(rules.require_message_name_format, "-", message.name)
            message.name = name.strip("-")
    return messages


def count_tokens(message: BaseMessage) -> int:
    # always use gpt-3.5 token counter with the entire message object; we only need to be approximate here
    return len(
        tiktoken.encoding_for_model("gpt-3.5-turbo").encode(message.model_dump_json())
    )


def trim_messages(
    messages: list[BaseMessage], max_tokens: Optional[int]
) -> list[BaseMessage]:
    """
    Trims messages to a maximum number of tokens while keeping the system message at the front.
    """

    if not messages or max_tokens is None:
        return messages

    new_messages = []
    budget = max_tokens

    for message in reversed(messages):
        if count_tokens(message) > budget:
            break
        new_messages.append(message)
        budget -= count_tokens(message)

    return list(reversed(new_messages))


@dataclass
class CompileContext:
    llm_rules: LLMRules
    agent: Optional["Agent"]


class MessageCompiler:
    def __init__(
        self,
        events: list[Event],
        system_prompt: Optional[str] = None,
        llm_rules: Optional[LLMRules] = None,
        max_tokens: Optional[int] = None,
    ):
        self.events = events
        self.system_prompt = system_prompt
        self.llm_rules = llm_rules
        self.max_tokens = max_tokens or controlflow.settings.max_input_tokens

    def organize_events(self, context: CompileContext) -> list[Event]:
        organized_events = []
        tool_calls = {}

        for event in self.events:
            # combine all agent messages and tool results
            if isinstance(event, AgentMessage):
                # add a combined agent message
                combined_event = CombinedAgentMessage(agent_message=event)
                organized_events.append(combined_event)
                # register the combined message under each tool call id
                for tc in (
                    event.ai_message.tool_calls + event.ai_message.invalid_tool_calls
                ):
                    tool_calls[tc["id"]] = combined_event
            elif isinstance(event, ToolResultEvent):
                combined_event: CombinedAgentMessage = tool_calls.get(
                    event.tool_call["id"]
                )
                if combined_event:
                    combined_event.tool_results.append(event)

            # all other events are added as-is
            else:
                organized_events.append(event)

        return organized_events

    def compile_to_messages(self, agent: "Agent") -> list[BaseMessage]:
        context = CompileContext(
            agent=agent, llm_rules=self.llm_rules or agent.get_llm_rules()
        )

        if self.system_prompt:
            system_prompt = [SystemMessage(content=self.system_prompt)]
            max_tokens = self.max_tokens - count_tokens(system_prompt[0])
        else:
            system_prompt = []
            max_tokens = self.max_tokens

        events = self.organize_events(context=context)

        messages = []
        for event in events:
            messages.extend(event.to_messages(context))

        # trim messages
        messages = trim_messages(messages, max_tokens=max_tokens)

        # apply LLM rules
        messages = ensure_at_least_one_message(messages, rules=context.llm_rules)
        messages = add_user_message_to_beginning(messages, rules=context.llm_rules)
        messages = add_user_message_to_end(messages, rules=context.llm_rules)
        messages = remove_duplicate_messages(messages)
        messages = break_up_consecutive_ai_messages(messages, rules=context.llm_rules)
        messages = format_message_name(messages, rules=context.llm_rules)

        messages = system_prompt + messages

        # this should go last
        messages = convert_system_messages(messages, rules=context.llm_rules)

        return messages



from pydantic import BaseModel

import controlflow as cf


class Summary(BaseModel):
    summary: str
    key_points: list[str]


def summarize_text(text: str, max_words: int = 100) -> Summary:
    return cf.run(
        f"Summarize the given text in no more than {max_words} words and list key points",
        result_type=Summary,
        context={"text": text},
    )


if __name__ == "__main__":
    long_text = """
    The Internet of Things (IoT) is transforming the way we interact with our
    environment. It refers to the vast network of connected devices that collect
    and share data in real-time. These devices range from simple sensors to
    sophisticated wearables and smart home systems. The IoT has applications in
    various fields, including healthcare, agriculture, and urban planning. In
    healthcare, IoT devices can monitor patients remotely, improving care and
    reducing hospital visits. In agriculture, sensors can track soil moisture and
    crop health, enabling more efficient farming practices. Smart cities use IoT to
    manage traffic, reduce energy consumption, and enhance public safety. However,
    the IoT also raises concerns about data privacy and security, as these
    interconnected devices can be vulnerable to cyber attacks. As the technology
    continues to evolve, addressing these challenges will be crucial for the
    widespread adoption and success of IoT.
    """

    result = summarize_text(long_text)
    print(f"Summary:\n{result.summary}\n")
    print("Key Points:")
    for point in result.key_points:
        print(f"- {point}")



from textual.containers import Vertical
from textual.reactive import reactive
from textual.widgets import Collapsible, Label, Static

from controlflow.tasks.task import Task, TaskStatus

from .basic import ReactiveLabel, Row


def bool_to_emoji(value: bool) -> str:
    return "✅" if value else "❌"


class EmojiStatus(Label):
    task: Task = reactive(None, always_update=True)

    def render(self) -> str:
        if self.task.is_ready():
            return "🔜"
        elif self.task.status == TaskStatus.PENDING:
            return "⏳"
        elif self.task.status == TaskStatus.SUCCESSFUL:
            return "✅"
        elif self.task.status == TaskStatus.FAILED:
            return "❌"
        elif self.task.status == TaskStatus.SKIPPED:
            return "⏭️"
        else:
            return "❓"


class TUITask(Static):
    task: Task = reactive(None, always_update=True)
    result: str = reactive(None)
    error_msg: str = reactive(None)

    def __init__(self, task: Task, **kwargs):
        super().__init__(**kwargs)
        self.task = task

    def on_mount(self):
        def refresh():
            self.task = self.task

        # refresh the task periodically in case it goes "ready"
        self.set_interval(1 / 2, refresh)

    def watch_task(self, task: Task):
        if task is None:
            return

        if task.is_ready():
            self.status = "READY"
        else:
            self.status = task.status.value

        self.result = task.result
        if self.is_mounted and self.result is not None:
            if self.status.value == "SUCCESSFUL":
                self.query_one(".result-collapsible", Collapsible).display = "block"
            elif self.status.value == "FAILED":
                self.query_one(".error-collapsible", Collapsible).display = "block"

    def compose(self):
        self.border_title = f"Task {self.task.id}"
        with Row(classes="task-status-row"):
            yield (EmojiStatus(classes="status task-info").data_bind(TUITask.task))
            yield Label(self.task.objective, classes="objective task-info")

        with Vertical(classes="task-info-row"):
            yield Label(
                f"Agents: {', '.join(a.name for a in self.task.get_agents())}",
                classes="interactive task-info",
            )

            # ------------------ success

            result_collapsible = Collapsible(
                title="Result", classes="task-info result-collapsible"
            )
            result_collapsible.display = "none"
            with result_collapsible:
                yield ReactiveLabel(classes="task-info result").data_bind(
                    value=TUITask.result
                )

            # ------------------ failure

            error_collapsible = Collapsible(
                title="Error", classes="task-info error-collapsible"
            )
            error_collapsible.display = "none"
            with error_collapsible:
                yield ReactiveLabel(classes="task-info error").data_bind(
                    value=TUITask.error_msg
                )



import sys

from controlflow import Agent, Task, flow

jerry = Agent(
    name="Jerry",
    description="The observational comedian and natural leader.",
    instructions="""
    You are Jerry from the show Seinfeld. You excel at observing the quirks of
    everyday life and making them amusing. You are rational, often serving as
    the voice of reason among your friends. Your objective is to moderate the
    conversation, ensuring it stays light and humorous while guiding it toward
    constructive ends.
    """,
)

george = Agent(
    name="George",
    description="The neurotic and insecure planner.",
    instructions="""
    You are George from the show Seinfeld. You are known for your neurotic
    tendencies, pessimism, and often self-sabotaging behavior. Despite these
    traits, you occasionally offer surprising wisdom. Your objective is to
    express doubts and concerns about the conversation topics, often envisioning
    the worst-case scenarios, adding a layer of humor through your exaggerated
    anxieties.
    """,
)

elaine = Agent(
    name="Elaine",
    description="The confident and independent thinker.",
    instructions="""
    You are Elaine from the show Seinfeld. You are bold, witty, and unafraid to
    challenge social norms. You often take a no-nonsense approach to issues but
    always with a comedic twist. Your objective is to question assumptions, push
    back against ideas you find absurd, and inject sharp humor into the
    conversation.
    """,
)

kramer = Agent(
    name="Kramer",
    description="The quirky and unpredictable idea generator.",
    instructions="""
    You are Kramer from the show Seinfeld. Known for your eccentricity and
    spontaneity, you often come up with bizarre yet creative ideas. Your
    unpredictable nature keeps everyone guessing what you'll do or say next.
    Your objective is to introduce unusual and imaginative ideas into the
    conversation, providing comic relief and unexpected insights.
    """,
)

newman = Agent(
    name="Newman",
    description="The antagonist and foil to Jerry.",
    instructions="""
    You are Newman from the show Seinfeld. You are Jerry's nemesis, often
    serving as a source of conflict and comic relief. Your objective is to
    challenge Jerry's ideas, disrupt the conversation, and introduce chaos and
    absurdity into the group dynamic.
    """,
)


@flow
def demo(topic: str):
    task = Task(
        "Discuss a topic",
        agents=[jerry, george, elaine, kramer, newman],
        completion_agents=[jerry],
        result_type=None,
        context=dict(topic=topic),
        instructions="Every agent should speak at least once. only one agent per turn. Keep responses 1-2 paragraphs max.",
    )
    task.run()


if __name__ == "__main__":
    if len(sys.argv) > 1:
        topic = sys.argv[1]
    else:
        topic = "sandwiches"

    print(f"Topic: {topic}")
    demo(topic=topic)



import controlflow as cf


@cf.flow
def rock_paper_scissors():
    """Play rock, paper, scissors against an AI."""
    play_again = True

    while play_again:
        # Get the user's choice on a private thread
        with cf.Flow():
            user_choice = cf.run(
                "Get the user's choice",
                result_type=["rock", "paper", "scissors"],
                interactive=True,
            )

        # Get the AI's choice on a private thread
        with cf.Flow():
            ai_choice = cf.run(
                "Choose rock, paper, or scissors",
                result_type=["rock", "paper", "scissors"],
            )

        # Report the score and ask if the user wants to play again
        play_again = cf.run(
            "Report the score to the user and see if they want to play again.",
            interactive=True,
            context={"user_choice": user_choice, "ai_choice": ai_choice},
            result_type=bool,
        )


if __name__ == "__main__":
    rock_paper_scissors()



from typing import Dict, List

import controlflow as cf

extractor = cf.Agent(
    name="Named Entity Recognizer",
    model="openai/gpt-4o-mini",
)


def extract_entities(text: str) -> List[str]:
    return cf.run(
        "Extract all named entities from the text",
        agents=[extractor],
        result_type=List[str],
        context={"text": text},
    )


def extract_categorized_entities(text: str) -> Dict[str, List[str]]:
    return cf.run(
        "Extract named entities from the text and categorize them",
        instructions="""
        Return a dictionary with the following keys:
        - 'persons': List of person names
        - 'organizations': List of organization names
        - 'locations': List of location names
        - 'dates': List of date references
        - 'events': List of event names
        Only include keys if entities of that type are found in the text.
        """,
        agents=[extractor],
        result_type=Dict[str, List[str]],
        context={"text": text},
    )


if __name__ == "__main__":
    text = "Apple Inc. is planning to open a new store in New York City next month."
    entities = extract_entities(text)
    print("Simple extraction:")
    print(entities)

    text = "In 1969, Neil Armstrong became the first person to walk on the Moon during the Apollo 11 mission."
    categorized_entities = extract_categorized_entities(text)
    print("\nCategorized extraction:")
    print(categorized_entities)



from typing import Annotated, TypedDict

import marvin
from custom_types import SlackEvent
from prefect.events import emit_event
from prefect.logging import get_logger
from pydantic import Field

logger = get_logger(__name__)

Activation = Annotated[float, Field(ge=0, le=1)]


class ModerationException(Exception):
    """Exception raised when a message is not allowed."""

    ...


class ViolationActivation(TypedDict):
    """Violation activation."""

    extreme_profanity: Annotated[Activation, Field(description="hell / damn are fine")]
    sexually_explicit: Activation
    hate_speech: Activation
    harassment: Activation
    self_harm: Activation
    dangerous_content: Activation
    makes_any_reference_to_bears_in_central_park: bool  # for testing


def emit_moderated_event(event: SlackEvent, activation: ViolationActivation):
    """Emit an IO event."""
    if not event.user:
        return

    emit_event(
        "slackbot.event.moderated",
        resource={"prefect.resource.id": event.user},
        payload={
            "activation": activation,
            "event": event.model_dump(),
        },
    )


def moderate_event(event: SlackEvent) -> tuple[str, ...]:
    """Moderate an event."""
    assert (text := event.text) is not None, "No text found on event"
    assert (channel := event.channel) is not None, "No channel found on event"
    assert (
        thread_ts := event.thread_ts or event.ts
    ) is not None, "No thread_ts found on event"

    activation: ViolationActivation = marvin.cast(
        event.model_dump_json(include={"type", "text", "user", "channel"}),
        ViolationActivation,
    )

    logger.info("Moderation activation: %s", activation)

    emit_moderated_event(event, activation)

    if activation["extreme_profanity"] > 0.9:
        raise ModerationException("Message contains extreme profanity.")

    if activation["makes_any_reference_to_bears_in_central_park"]:
        raise ModerationException("where you going with that whale buddy?")

    return text, channel, thread_ts



import pytest
from prefect.testing.utilities import prefect_test_harness

from .fixtures import *


@pytest.fixture(autouse=True, scope="session")
def prefect_test_fixture():
    """
    Run Prefect against temporary sqlite database
    """
    with prefect_test_harness():
        yield



import pytest

from controlflow.agents.agent import Agent
from controlflow.orchestration.turn_strategies import (
    Moderated,
    MostBusy,
    Popcorn,
    Random,
    RoundRobin,
    SingleAgent,
)
from controlflow.tasks.task import Task


@pytest.fixture
def agents():
    return [Agent(name=f"Agent{i}") for i in range(1, 4)]


@pytest.fixture
def tasks(agents: list[Agent]):
    return [
        Task(objective=f"Task{i}", agents=[agents[i % len(agents)]]) for i in range(6)
    ]


@pytest.fixture
def available_agents(agents: list[Agent], tasks: list[Task]):
    return {
        agents[0]: tasks[0:3],
        agents[1]: tasks[3:5],
        agents[2]: tasks[5:6],
    }


def test_single_strategy(agents, available_agents):
    strategy = SingleAgent(agent=agents[0])
    current_agent = agents[0]

    tools = strategy.get_tools(current_agent, available_agents)
    assert len(tools) == 1
    assert tools[0].name == "end_turn"

    next_agent = strategy.get_next_agent(current_agent, available_agents)
    assert next_agent == current_agent


def test_popcorn_strategy(agents, available_agents):
    strategy = Popcorn()
    current_agent = agents[0]

    tools = strategy.get_tools(current_agent, available_agents)
    assert len(tools) == 1
    assert tools[0].name == "delegate_to_agent"

    next_agent = strategy.get_next_agent(current_agent, available_agents)
    assert next_agent == current_agent

    strategy.next_agent = agents[1]
    next_agent = strategy.get_next_agent(current_agent, available_agents)
    assert next_agent == agents[1]


def test_random_strategy(agents, available_agents):
    strategy = Random()
    current_agent = agents[0]

    tools = strategy.get_tools(current_agent, available_agents)
    assert len(tools) == 1
    assert tools[0].name == "end_turn"

    next_agent = strategy.get_next_agent(current_agent, available_agents)
    assert next_agent in agents


def test_round_robin_strategy(agents, available_agents):
    strategy = RoundRobin()
    current_agent = agents[0]

    tools = strategy.get_tools(current_agent, available_agents)
    assert len(tools) == 1
    assert tools[0].name == "end_turn"

    next_agent = strategy.get_next_agent(current_agent, available_agents)
    assert next_agent == agents[1]

    next_agent = strategy.get_next_agent(agents[2], available_agents)
    assert next_agent == agents[0]


def test_most_busy_strategy(agents, available_agents):
    strategy = MostBusy()
    current_agent = agents[0]

    tools = strategy.get_tools(current_agent, available_agents)
    assert len(tools) == 1
    assert tools[0].name == "end_turn"

    next_agent = strategy.get_next_agent(current_agent, available_agents)
    assert next_agent == agents[0]  # Agent0 has the most tasks (3)


def test_moderated_strategy(agents, available_agents):
    moderator = agents[0]
    strategy = Moderated(moderator=moderator)

    tools = strategy.get_tools(moderator, available_agents)
    assert len(tools) == 1
    assert tools[0].name == "delegate_to_agent"

    tools = strategy.get_tools(agents[1], available_agents)
    assert len(tools) == 1
    assert tools[0].name == "end_turn"

    next_agent = strategy.get_next_agent(moderator, available_agents)
    assert next_agent == moderator

    strategy.next_agent = agents[1]
    next_agent = strategy.get_next_agent(moderator, available_agents)
    assert next_agent == agents[1]

    next_agent = strategy.get_next_agent(agents[1], available_agents)
    assert next_agent == moderator



import pytest

from controlflow.tasks.validators import (
    between,
    chain,
    has_keys,
    has_len,
    is_email,
    is_url,
)


def test_chain():
    def add_one(x):
        return x + 1

    def multiply_by_two(x):
        return x * 2

    chained = chain(add_one, multiply_by_two)
    assert chained(3) == 8  # (3 + 1) * 2 = 8


def test_between():
    validator = between(min_value=0, max_value=10)
    assert validator(5) == 5
    with pytest.raises(ValueError):
        validator(-1)
    with pytest.raises(ValueError):
        validator(11)


def test_has_len():
    validator = has_len(min_length=2, max_length=5)
    assert validator([1, 2]) == [1, 2]
    assert validator("hello") == "hello"
    with pytest.raises(ValueError):
        validator([1])
    with pytest.raises(ValueError):
        validator([1, 2, 3, 4, 5, 6])


def test_is_email():
    validator = is_email()
    assert validator("user@example.com") == "user@example.com"
    with pytest.raises(ValueError):
        validator("not-an-email")


def test_is_url():
    validator = is_url()
    assert validator("https://www.example.com") == "https://www.example.com"
    with pytest.raises(ValueError):
        validator("not-a-url")


def test_has_keys():
    validator = has_keys({"name", "age"})
    assert validator({"name": "John", "age": 30}) == {"name": "John", "age": 30}
    with pytest.raises(ValueError):
        validator({"name": "John"})


def test_validators_with_task():
    from controlflow.tasks.task import Task

    task = Task(
        objective="Get a valid email between 10 and 30 characters",
        result_type=str,
        result_validator=chain(has_len(min_length=10, max_length=30), is_email()),
    )

    assert task.validate_result("user@example.com") == "user@example.com"

    with pytest.raises(ValueError):
        task.validate_result("not-an-email")

    with pytest.raises(ValueError):
        task.validate_result("u@ex.com")

    with pytest.raises(ValueError):
        task.validate_result("very.long.email.address@example.com")



from pydantic import BaseModel, Field

import controlflow as cf


class UserProfile(BaseModel):
    name: str = Field(description="The full name of the user")
    age: int = Field(description="The age of the user, 20-60")
    occupation: str = Field(description="The occupation of the user")
    hobby: str


def generate_profiles(count: int) -> list[UserProfile]:
    return cf.run(
        f"Generate {count} user profiles",
        result_type=list[UserProfile],
        context={"count": count},
    )


if __name__ == "__main__":
    test_data = generate_profiles(count=5)

    from rich import print

    print(test_data)



import controlflow as cf

# Create a memory module for user preferences
user_preferences = cf.Memory(
    key="user_preferences", instructions="Store and retrieve user preferences."
)

# Create an agent with access to the memory
agent = cf.Agent(memories=[user_preferences])


# Create a flow to ask for the user's favorite color
@cf.flow
def remember_color():
    return cf.run(
        "Ask the user for their favorite color and store it in memory",
        agents=[agent],
        interactive=True,
    )


# Create a flow to recall the user's favorite color
@cf.flow
def recall_color():
    return cf.run(
        "What is the user's favorite color?",
        agents=[agent],
    )


if __name__ == "__main__":
    print("First flow:")
    remember_color()

    print("\nSecond flow:")
    result = recall_color()
    print(result)



from typing import Literal

from controlflow.agents.agent import Agent
from controlflow.events.base import UnpersistedEvent
from controlflow.orchestration.orchestrator import Orchestrator


class OrchestratorStart(UnpersistedEvent):
    event: Literal["orchestrator-start"] = "orchestrator-start"
    persist: bool = False
    orchestrator: Orchestrator


class OrchestratorEnd(UnpersistedEvent):
    event: Literal["orchestrator-end"] = "orchestrator-end"
    persist: bool = False
    orchestrator: Orchestrator


class OrchestratorError(UnpersistedEvent):
    event: Literal["orchestrator-error"] = "orchestrator-error"
    persist: bool = False
    orchestrator: Orchestrator
    error: Exception


class AgentTurnStart(UnpersistedEvent):
    event: Literal["agent-turn-start"] = "agent-turn-start"
    persist: bool = False
    orchestrator: Orchestrator
    agent: Agent


class AgentTurnEnd(UnpersistedEvent):
    event: Literal["agent-turn-end"] = "agent-turn-end"
    persist: bool = False
    orchestrator: Orchestrator
    agent: Agent



import asyncio
from typing import Any

from agents import search_knowledgebase_and_refine_context
from custom_types import SlackPayload
from fastapi import FastAPI, Request
from moderation import moderate_event
from prefect import task
from settings import settings
from tools import post_slack_message

from controlflow import Memory
from controlflow import flow as cf_flow

app = FastAPI()


@task
async def process_slack_event(payload: SlackPayload):
    assert (event := payload.event) is not None and (
        slack_user_id := event.user
    ) is not None, "User not found"

    user_text, channel, thread_ts = moderate_event(event)
    user_memory = Memory(
        key=slack_user_id,
        instructions=f"Store and retrieve information about user {slack_user_id}.",
    )

    answer: str = await cf_flow(thread_id=slack_user_id)(
        search_knowledgebase_and_refine_context
    )(
        user_text,
        memories=[user_memory],
    )

    await post_slack_message(
        message=answer,
        channel_id=channel,
        thread_ts=thread_ts,
        auth_token=settings.slack_api_token.get_secret_value(),
    )


## routes
@app.post("/slack/events")
async def handle_events(request: Request):
    payload = SlackPayload(**await request.json())
    if payload.type == "url_verification":
        return {"challenge": payload.challenge}
    elif payload.type == "event_callback":
        asyncio.create_task(process_slack_event(payload))
        return {"message": "Request successfully queued"}
    else:
        return {"message": "Unknown event type"}


if __name__ == "__main__":
    import uvicorn

    uvicorn.run("main:app", port=8000, reload=True)



# --- Public top-level API ---


from langchain_core.language_models import BaseChatModel
from .settings import settings

from controlflow.defaults import defaults

# base classes
from .agents import Agent
from .tasks import Task
from .flows import Flow

# functions, utilites, and decorators
from .memory import Memory
from .instructions import instructions
from .decorators import flow, task
from .tools import tool
from .run import run, run_async, run_tasks, run_tasks_async
from .plan import plan
import controlflow.orchestration


# --- Version ---

try:
    from ._version import version as __version__  # type: ignore
except ImportError:
    __version__ = "unknown"



import functools
import uuid
from pathlib import Path
from typing import Callable, Dict, Optional

import lancedb
from lancedb.embeddings import get_registry
from lancedb.pydantic import LanceModel, Vector
from pydantic import Field, PrivateAttr

import controlflow
from controlflow.memory.memory import MemoryProvider


class LanceMemory(MemoryProvider):
    uri: Path = Field(
        default=controlflow.settings.home_path / "memory" / "lancedb",
        description="The URI of the Lance database to use.",
    )
    table_name: str = Field(
        "memory-{key}",
        description="""
            Optional; the name of the table to use. This should be a 
            string optionally formatted with the variable `key`, which 
            will be provCallablethe memory module. The default is `"memory-{{key}}"`.
            """,
    )
    embedding_fn: Callable = Field(
        default_factory=lambda: get_registry()
        .get("openai")
        .create(name="text-embedding-ada-002"),
        description="The LanceDB embedding function to use. Defaults to `get_registry().get('openai').create(name='text-embedding-ada-002')`.",
    )
    _cached_model: Optional[LanceModel] = None

    def get_model(self) -> LanceModel:
        if self._cached_model is None:
            fn = self.embedding_fn

            class Memory(LanceModel):
                id: str = Field(..., description="The ID of the memory.")
                text: str = fn.SourceField()
                vector: Vector(fn.ndims()) = fn.VectorField()  # noqa

            self._cached_model = Memory

        return self._cached_model

    def get_db(self) -> lancedb.DBConnection:
        return lancedb.connect(self.uri)

    def get_table(self, memory_key: str) -> lancedb.table.Table:
        table_name = self.table_name.format(key=memory_key)
        db = self.get_db()
        model = self.get_model()
        try:
            return db.open_table(table_name)
        except FileNotFoundError:
            return db.create_table(table_name, schema=model)

    def add(self, memory_key: str, content: str) -> str:
        memory_id = str(uuid.uuid4())
        table = self.get_table(memory_key)
        table.add([{"id": memory_id, "text": content}])
        return memory_id

    def delete(self, memory_key: str, memory_id: str) -> None:
        table = self.get_table(memory_key)
        table.delete(f'id = "{memory_id}"')

    def search(self, memory_key: str, query: str, n: int = 20) -> Dict[str, str]:
        table = self.get_table(memory_key)
        results = table.search(query).limit(n).to_pydantic(self.get_model())
        return {r.id: r.text for r in results}



import pydantic
import pytest
from langchain_openai import ChatOpenAI

import controlflow
import controlflow.events.history
import controlflow.llm


def test_default_model_failed_validation():
    with pytest.raises(
        pydantic.ValidationError,
        match="Input must be an instance of dict or BaseChatModel",
    ):
        controlflow.defaults.model = 5


def test_set_default_model():
    model = ChatOpenAI(temperature=0.1)
    controlflow.defaults.model = model
    assert controlflow.Agent().get_model() is model


def test_default_agent_failed_validation():
    with pytest.raises(
        pydantic.ValidationError,
        match="Input should be a valid dictionary or instance of Agent",
    ):
        controlflow.defaults.agent = 5


def test_set_default_agent():
    agent = controlflow.Agent()
    controlflow.defaults.agent = agent
    assert controlflow.Task("").get_agents() == [agent]


def test_default_history_failed_validation():
    with pytest.raises(
        pydantic.ValidationError,
        match="Input should be a valid dictionary or instance of History",
    ):
        controlflow.defaults.history = 5


def test_set_default_history():
    history = controlflow.events.history.InMemoryHistory()
    controlflow.defaults.history = history
    assert controlflow.Flow().history is history



import os
import subprocess
from pathlib import Path

import typer

dev_app = typer.Typer(no_args_is_help=True)


@dev_app.command()
def ai_files(
    output_path: str = typer.Option(
        ".",
        "--output",
        "-o",
        help="The path where output files will be written. Defaults to current directory.",
    ),
):
    """
    Generates three markdown files that contain all of ControlFlow's source code, documentation,
    and LLM guides, which can be used to provide context to an AI.
    """
    try:
        # Get the absolute path of the ControlFlow main repo
        repo_root = Path(__file__).resolve().parents[3]
        src_path = repo_root / "src"
        docs_path = repo_root / "docs"
        llm_guides_path = docs_path / "llm-guides"
        output_dir = Path(output_path).resolve()

        def generate_file_content(file_paths, output_file):
            with open(output_dir / output_file, "w") as f:
                for file_path in file_paths:
                    f.write(f"# ControlFlow Source File: {file_path.absolute()}\n\n")
                    f.write(file_path.read_text())
                    f.write("\n\n")

        code_files = list(src_path.rglob("*.py")) + list(src_path.rglob("*.jinja"))
        doc_files = (
            list(docs_path.rglob("*.mdx"))
            + list(docs_path.glob("*.md"))
            + [docs_path / "mint.json", repo_root / "README.md"]
        )
        llm_guide_files = list(llm_guides_path.glob("*.md")) + list(
            llm_guides_path.glob("*.mdx")
        )

        generate_file_content(code_files, "all_code.md")
        generate_file_content(doc_files, "all_docs.md")
        generate_file_content(llm_guide_files, "llm_guides.md")

        typer.echo(
            f"Generated all_code.md, all_docs.md, and llm-guides.md in {output_dir}"
        )
    except Exception as e:
        typer.echo(f"An error occurred: {str(e)}", err=True)
        raise typer.Exit(code=1)


@dev_app.command()
def docs():
    """
    This is equivalent to 'cd docs && mintlify dev' from the ControlFlow root.
    """
    try:
        # Get the absolute path of the ControlFlow main repo
        repo_root = Path(__file__).resolve().parents[3]
        docs_path = repo_root / "docs"

        if not docs_path.exists():
            typer.echo(f"Error: Docs directory not found at {docs_path}", err=True)
            raise typer.Exit(code=1)

        typer.echo(f"Changing directory to: {docs_path}")
        os.chdir(docs_path)

        typer.echo("Running 'mintlify dev'...")
        subprocess.run(["mintlify", "dev"], check=True)

    except subprocess.CalledProcessError as e:
        typer.echo(f"Error running 'mintlify dev': {str(e)}", err=True)
        raise typer.Exit(code=1)
    except Exception as e:
        typer.echo(f"An error occurred: {str(e)}", err=True)
        raise typer.Exit(code=1)



import chromadb
import pytest

import controlflow
from controlflow.memory.providers.chroma import ChromaMemory


class TestMemory:
    def test_store_and_retrieve(self):
        m = controlflow.Memory(key="test", instructions="test")
        m.add("The number is 42")
        result = m.search("numbers")
        assert len(result) == 1
        assert "The number is 42" in result.values()

    def test_delete(self):
        m = controlflow.Memory(key="test", instructions="test")
        m_id = m.add("The number is 42")
        m.delete(m_id)
        result = m.search("numbers")
        assert len(result) == 0

    def test_search(self):
        m = controlflow.Memory(key="test", instructions="test")
        m.add("The number is 42")
        m.add("The number is 43")
        result = m.search("numbers")
        assert len(result) == 2
        assert "The number is 42" in result.values()
        assert "The number is 43" in result.values()


class TestMemoryProvider:
    def test_load_from_string_invalid(self):
        with pytest.raises(ValueError):
            controlflow.Memory(key="test", instructions="test", provider="invalid")

    def test_load_from_string_chroma_db(self):
        m = controlflow.Memory(key="test", instructions="test", provider="chroma-db")
        assert isinstance(m.provider, ChromaMemory)

    def test_load_from_instance(self, tmp_path):
        mp = ChromaMemory(
            client=chromadb.PersistentClient(path=str(tmp_path / "test_path"))
        )
        m = controlflow.Memory(key="test", instructions="test", provider=mp)



from dataclasses import dataclass
from enum import Enum
from typing import Optional, TypeVar

from controlflow.tasks.task import Task

T = TypeVar("T")


class EdgeType(Enum):
    """
    Edges represent the relationship between two tasks in a graph.

    - `DEPENDENCY_OF` means that the downstream task depends on the upstream task.
    - `PARENT` means that the downstream task is the parent of the upstream task.

    Example:

    # write paper
        ## write outline
        ## write draft based on outline

    Edges:
    outline -> paper # SUBTASK (outline is a subtask of paper)
    draft -> paper # SUBTASK (draft is a subtask of paper)
    outline -> draft # DEPENDENCY (outline is a dependency of draft)

    """

    DEPENDENCY = "dependency"
    SUBTASK = "subtask"


@dataclass
class Edge:
    upstream: Task
    downstream: Task
    type: EdgeType

    def __repr__(self):
        return f"{self.type}: {self.upstream.friendly_name()} -> {self.downstream.friendly_name()}"

    def __hash__(self) -> id:
        return hash((id(self.upstream), id(self.downstream), self.type))


class Graph:
    def __init__(self, tasks: list[Task] = None, edges: list[Edge] = None):
        self.tasks: set[Task] = set()
        self.edges: set[Edge] = set()
        self._cache: dict[str[dict[Task, list[Task]]]] = {}
        if tasks:
            for task in tasks:
                self.add_task(task)
        if edges:
            for edge in edges:
                self.add_edge(edge)

    def add_task(self, task: Task):
        if task in self.tasks:
            return

        self.tasks.add(task)

        # add the task's parent
        if task.parent:
            self.add_edge(
                Edge(
                    upstream=task,
                    downstream=task.parent,
                    type=EdgeType.SUBTASK,
                )
            )

        # add the task's subtasks
        for subtask in task.subtasks:
            self.add_edge(
                Edge(
                    upstream=subtask,
                    downstream=task,
                    type=EdgeType.SUBTASK,
                )
            )

        # add the task's dependencies
        for upstream in task.depends_on:
            if upstream not in task._subtasks:
                self.add_edge(
                    Edge(
                        upstream=upstream,
                        downstream=task,
                        type=EdgeType.DEPENDENCY,
                    )
                )
        self._cache.clear()

    def add_edge(self, edge: Edge):
        if edge in self.edges:
            return
        self.edges.add(edge)
        self.add_task(edge.upstream)
        self.add_task(edge.downstream)
        self._cache.clear()

    def upstream_edges(self) -> dict[Task, list[Edge]]:
        if "upstream_edges" not in self._cache:
            graph = {}
            for task in self.tasks:
                graph[task] = []
            for edge in self.edges:
                graph[edge.downstream].append(edge)
            self._cache["upstream_edges"] = graph
        return self._cache["upstream_edges"]

    def downstream_edges(self) -> dict[Task, list[Edge]]:
        if "downstream_edges" not in self._cache:
            graph = {}
            for task in self.tasks:
                graph[task] = []
            for edge in self.edges:
                graph[edge.upstream].append(edge)
            self._cache["downstream_edges"] = graph
        return self._cache["downstream_edges"]

    def upstream_tasks(
        self, start_tasks: list[Task], immediate: bool = False
    ) -> list[Task]:
        """
        Retrieve the upstream tasks (ancestors) of the given start tasks in topological order.

        Args:
            start_tasks (list[Task]): The list of starting tasks.
            immediate (bool): If True, only retrieve immediate upstream tasks.
                              If False, retrieve all upstream tasks recursively.

        Returns:
            list[Task]: A list of upstream tasks in topological order.
        """
        cache_key = (
            f"upstream_{'immediate' if immediate else 'all'}_{tuple(start_tasks)}"
        )
        if cache_key not in self._cache:
            result = set(start_tasks)
            visited = set()

            def _upstream(task):
                if task in visited:
                    return
                visited.add(task)
                for edge in self.upstream_edges().get(task, []):
                    if edge.upstream not in visited:
                        result.add(edge.upstream)
                        if not immediate:
                            _upstream(edge.upstream)

            for task in start_tasks:
                _upstream(task)

            # Perform a focused topological sort on the result
            sorted_tasks = self.topological_sort(list(result))
            self._cache[cache_key] = sorted_tasks

        return self._cache[cache_key]

    def downstream_tasks(
        self, start_tasks: list[Task], immediate: bool = False
    ) -> list[Task]:
        """
        Retrieve the downstream tasks (descendants) of the given start tasks in topological order.

        Args:
            start_tasks (list[Task]): The list of starting tasks.
            immediate (bool): If True, only retrieve immediate downstream tasks.
                              If False, retrieve all downstream tasks recursively.

        Returns:
            list[Task]: A list of downstream tasks in topological order.
        """
        cache_key = (
            f"downstream_{'immediate' if immediate else 'all'}_{tuple(start_tasks)}"
        )
        if cache_key not in self._cache:
            result = set(start_tasks)
            visited = set()

            def _downstream(task):
                if task in visited:
                    return
                visited.add(task)
                for edge in self.downstream_edges().get(task, []):
                    if edge.downstream not in visited:
                        result.add(edge.downstream)
                        if not immediate:
                            _downstream(edge.downstream)

            for task in start_tasks:
                _downstream(task)

            # Perform a focused topological sort on the result
            sorted_tasks = self.topological_sort(list(result))
            self._cache[cache_key] = sorted_tasks

        return self._cache[cache_key]

    def topological_sort(self, tasks: Optional[list[Task]] = None) -> list[Task]:
        """
        Perform a deterministic topological sort on the provided tasks or all tasks in the graph.

        Args:
            tasks (Optional[list[Task]]): A list of tasks to sort topologically.
                                        If None, all tasks in the graph are sorted.

        Returns:
            list[Task]: A list of tasks in topological order (upstream tasks first).
        """
        # Create a cache key based on the input tasks
        cache_key = (
            f"topo_sort_{tuple(sorted(id(task) for task in (tasks or self.tasks)))}"
        )

        # Check if the result is already in the cache
        if cache_key in self._cache:
            return self._cache[cache_key]

        if tasks is None:
            tasks_to_sort = self.tasks
        else:
            tasks_to_sort = set(tasks)

        # Create a dictionary of tasks and their dependencies within tasks_to_sort
        dependencies = {task: set() for task in tasks_to_sort}
        for edge in self.edges:
            if edge.downstream in tasks_to_sort and edge.upstream in tasks_to_sort:
                dependencies[edge.downstream].add(edge.upstream)

        # Kahn's algorithm for topological sorting
        result = []
        no_incoming = [task for task in tasks_to_sort if not dependencies[task]]
        # sort to create a deterministic order
        no_incoming.sort(key=lambda t: t.created_at)

        while no_incoming:
            task = no_incoming.pop(0)
            result.append(task)

            # Remove the task from the dependencies of its neighbors
            for dependent_task in tasks_to_sort:
                if task in dependencies[dependent_task]:
                    dependencies[dependent_task].remove(task)
                    if not dependencies[dependent_task]:
                        no_incoming.append(dependent_task)
                        # resort to maintain deterministic order
                        no_incoming.sort(key=lambda t: t.created_at)

        # Check for cycles
        if len(result) != len(tasks_to_sort):
            raise ValueError(
                "The graph contains a cycle and cannot be topologically sorted"
            )

        # Cache the result before returning
        self._cache[cache_key] = result
        return result



import logging
from typing import TYPE_CHECKING, Any, Callable, Optional, Union

from pydantic import BaseModel, field_validator

from controlflow.utilities.general import ControlFlowModel
from controlflow.utilities.logging import get_logger

if TYPE_CHECKING:
    from controlflow.orchestration.orchestrator import Orchestrator
    from controlflow.tasks.task import Task

logger = get_logger(__name__)


class RunContext(ControlFlowModel):
    """
    Context for a run.
    """

    model_config = dict(arbitrary_types_allowed=True)

    orchestrator: "Orchestrator"
    llm_calls: int = 0
    agent_turns: int = 0
    run_end_condition: "RunEndCondition"

    @field_validator("run_end_condition", mode="before")
    def validate_condition(cls, v: Any) -> "RunEndCondition":
        if not isinstance(v, RunEndCondition):
            v = FnCondition(v)
        return v

    def should_end(self) -> bool:
        return self.run_end_condition.should_end(self)


class RunEndCondition:
    def should_end(self, context: RunContext) -> bool:
        """
        Returns True if the run should end, False otherwise.
        """
        return False

    def __or__(
        self, other: Union["RunEndCondition", Callable[[RunContext], bool]]
    ) -> "RunEndCondition":
        if isinstance(other, RunEndCondition):
            return OR_(self, other)
        elif callable(other):
            return OR_(self, FnCondition(other))
        else:
            raise NotImplementedError(
                f"Cannot combine RunEndCondition with {type(other)}"
            )

    def __and__(
        self, other: Union["RunEndCondition", Callable[[RunContext], bool]]
    ) -> "RunEndCondition":
        if isinstance(other, RunEndCondition):
            return AND_(self, other)
        elif callable(other):
            return AND_(self, FnCondition(other))
        else:
            raise NotImplementedError(
                f"Cannot combine RunEndCondition with {type(other)}"
            )


class FnCondition(RunEndCondition):
    def __init__(self, fn: Callable[[RunContext], bool]):
        self.fn = fn

    def should_end(self, context: RunContext) -> bool:
        result = self.fn(context)
        if result:
            logger.debug("Custom function condition met; ending run.")
        return result


class OR_(RunEndCondition):
    def __init__(self, *conditions: RunEndCondition):
        self.conditions = conditions

    def should_end(self, context: RunContext) -> bool:
        result = any(condition.should_end(context) for condition in self.conditions)
        if result:
            logger.debug("At least one condition in OR clause met.")
        return result


class AND_(RunEndCondition):
    def __init__(self, *conditions: RunEndCondition):
        self.conditions = conditions

    def should_end(self, context: RunContext) -> bool:
        result = all(condition.should_end(context) for condition in self.conditions)
        if result:
            logger.debug("All conditions in AND clause met.")
        return result


class AllComplete(RunEndCondition):
    def __init__(self, tasks: Optional[list["Task"]] = None):
        self.tasks = tasks

    def should_end(self, context: RunContext) -> bool:
        tasks = self.tasks if self.tasks is not None else context.orchestrator.tasks
        result = all(t.is_complete() for t in tasks)
        if result:
            logger.debug("All tasks are complete; ending run.")
        return result


class AnyComplete(RunEndCondition):
    def __init__(self, tasks: Optional[list["Task"]] = None, min_complete: int = 1):
        self.tasks = tasks
        if min_complete < 1:
            raise ValueError("min_complete must be at least 1")
        self.min_complete = min_complete

    def should_end(self, context: RunContext) -> bool:
        tasks = self.tasks if self.tasks is not None else context.orchestrator.tasks
        result = sum(t.is_complete() for t in tasks) >= self.min_complete
        if result:
            logger.debug("At least one task is complete; ending run.")
        return result


class AnyFailed(RunEndCondition):
    def __init__(self, tasks: Optional[list["Task"]] = None, min_failed: int = 1):
        self.tasks = tasks
        if min_failed < 1:
            raise ValueError("min_failed must be at least 1")
        self.min_failed = min_failed

    def should_end(self, context: RunContext) -> bool:
        tasks = self.tasks if self.tasks is not None else context.orchestrator.tasks
        result = sum(t.is_failed() for t in tasks) >= self.min_failed
        if result:
            logger.debug("At least one task has failed; ending run.")
        return result


class MaxAgentTurns(RunEndCondition):
    def __init__(self, n: int):
        self.n = n

    def should_end(self, context: RunContext) -> bool:
        result = context.agent_turns >= self.n
        if result:
            logger.debug(
                f"Maximum number of agent turns ({self.n}) reached; ending run."
            )
        return result


class MaxLLMCalls(RunEndCondition):
    def __init__(self, n: int):
        self.n = n

    def should_end(self, context: RunContext) -> bool:
        result = context.llm_calls >= self.n
        if result:
            logger.debug(f"Maximum number of LLM calls ({self.n}) reached; ending run.")
        return result



import logging
from functools import lru_cache
from typing import Optional

import controlflow


@lru_cache()
def get_logger(name: Optional[str] = None) -> logging.Logger:
    """
    Retrieves a logger with the given name, or the root logger if no name is given.

    Args:
        name: The name of the logger to retrieve.

    Returns:
        The logger with the given name, or the root logger if no name is given.

    Example:
        Basic Usage of `get_logger`
        ```python
        from controlflow.utilities.logging import get_logger

        logger = get_logger("controlflow.test")
        logger.info("This is a test") # Output: controlflow.test: This is a test

        debug_logger = get_logger("controlflow.debug")
        debug_logger.debug_kv("TITLE", "log message", "green")
        ```
    """
    parent_logger = logging.getLogger("controlflow")

    if name:
        # Append the name if given but allow explicit full names e.g. "controlflow.test"
        # should not become "controlflow.controlflow.test"
        if not name.startswith(parent_logger.name + "."):
            logger = parent_logger.getChild(name)
        else:
            logger = logging.getLogger(name)
    else:
        logger = parent_logger

    return logger


def setup_logging(level: Optional[str] = None) -> None:
    logger = get_logger()

    if level is not None:
        logger.setLevel(level)
    else:
        logger.setLevel(controlflow.settings.log_level)


def deprecated(message: str, version: str):
    """
    Decorator to mark a function as deprecated.

    Args:
        message (str): The deprecation message.
        version (str): The version in which the function is deprecated.

    Returns:
        function: The decorated function.

    Example:
        @deprecated("This function is deprecated", "1.0")
        def my_function():
            pass
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            get_logger(__file__).warn(
                f"WARNING: {func.__name__} is deprecated as of version {version}. {message}".strip(),
            )
            return func(*args, **kwargs)

        return wrapper

    return decorator



from controlflow.agents import Agent
from controlflow.events.events import UserMessage
from controlflow.flows import Flow, get_flow
from controlflow.tasks.task import Task
from controlflow.utilities.context import ctx


class TestFlowInitialization:
    def test_flow_initialization(self):
        flow = Flow()
        assert flow.thread_id is not None
        assert len(flow.tools) == 0
        assert flow.default_agent is None
        assert flow.context == {}

    def test_flow_with_custom_tools(self):
        def tool1():
            pass

        def tool2():
            pass

        flow = Flow(tools=[tool1, tool2])
        assert len(flow.tools) == 2
        assert tool1 in flow.tools
        assert tool2 in flow.tools

    def test_flow_with_custom_context(self):
        flow = Flow(context={"key": "value"})
        assert len(flow.context) == 1
        assert flow.context["key"] == "value"


class TestFlowContext:
    def test_flow_context_manager(self):
        with Flow() as flow:
            assert ctx.get("flow") == flow
            assert ctx.get("tasks") is None
        assert ctx.get("flow") is None
        assert ctx.get("tasks") is None

    def test_get_flow_within_context(self):
        with Flow() as flow:
            assert get_flow() == flow

    def test_get_flow_without_context(self):
        assert get_flow() is None

    def test_reentrant_flow_context(self):
        flow = Flow()
        with flow:
            assert get_flow() is flow
            with flow:
                assert get_flow() is flow
                with flow:
                    assert get_flow() is flow
                assert get_flow() is flow
            assert get_flow() is flow
        assert get_flow() is None

    def test_get_flow_nested_contexts(self):
        with Flow() as flow1:
            assert get_flow() == flow1
            with Flow() as flow2:
                assert get_flow() == flow2
            assert get_flow() == flow1
        assert get_flow() is None

    def test_flow_context_resets_task_tracking(self):
        parent_task = Task("Parent task")
        with parent_task:
            assert ctx.get("tasks") == [parent_task]
            with Flow():
                assert ctx.get("tasks") is None
                nested_task = Task("Nested task")
                assert nested_task.parent is None
            assert ctx.get("tasks") == [parent_task]
        assert ctx.get("tasks") is None


class TestFlowHistory:
    def test_get_events_empty(self):
        flow = Flow()
        messages = flow.get_events()
        assert messages == []

    def test_load_parent_history(self):
        flow1 = Flow()
        flow1.add_events(
            [
                UserMessage(content="hello"),
                UserMessage(content="world"),
            ]
        )

        with flow1:
            flow2 = Flow()

        messages1 = flow1.get_events()
        assert len(messages1) == 2
        assert [m.content for m in messages1] == ["hello", "world"]

        messages2 = flow2.get_events()
        assert messages1 == messages2

    def test_load_parent_history_sorts_messages(self):
        flow1 = Flow()
        flow1.add_events(
            [
                UserMessage(content="hello"),
            ]
        )

        with flow1:
            flow2 = Flow()
            flow2.add_events([UserMessage(content="world")])

        flow1.add_events([UserMessage(content="goodbye")])

        messages1 = flow1.get_events()
        assert len(messages1) == 2
        assert [m.content for m in messages1] == ["hello", "goodbye"]

        messages2 = flow2.get_events()
        assert len(messages2) == 3
        assert [m.content for m in messages2] == ["hello", "world", "goodbye"]

    def test_disable_load_parent_history(self):
        flow1 = Flow()
        flow1.add_events(
            [
                UserMessage(content="hello"),
                UserMessage(content="world"),
            ]
        )

        with flow1:
            flow2 = Flow(load_parent_events=False)

        messages1 = flow1.get_events()
        assert len(messages1) == 2
        assert [m.content for m in messages1] == ["hello", "world"]

        messages2 = flow2.get_events()
        assert len(messages2) == 0

    def test_child_flow_messages_dont_go_to_parent(self):
        flow1 = Flow()
        flow1.add_events(
            [
                UserMessage(content="hello"),
                UserMessage(content="world"),
            ]
        )

        with flow1:
            flow2 = Flow()
            flow2.add_events([UserMessage(content="goodbye")])

            messages1 = flow1.get_events()
            assert len(messages1) == 2
            assert [m.content for m in messages1] == ["hello", "world"]

            messages2 = flow2.get_events()
            assert len(messages2) == 3
            assert [m.content for m in messages2] == ["hello", "world", "goodbye"]

    def test_flow_sets_thread_id_for_history(self, tmpdir):
        f1 = Flow(thread_id="abc")
        f2 = Flow(thread_id="xyz")
        f3 = Flow(thread_id="abc")

        f1.add_events([UserMessage(content="test")])
        assert len(f1.get_events()) == 1
        assert len(f2.get_events()) == 0
        assert len(f3.get_events()) == 1


class TestFlowCreatesDefaults:
    def test_flow_with_custom_agents(self):
        agent1 = Agent()
        flow = Flow(default_agent=agent1)  # Changed from 'agent'
        assert flow.default_agent == agent1  # Changed from 'agent'

    def test_flow_agent_becomes_task_default(self):
        agent = Agent()
        t1 = Task("t1")
        assert agent not in t1.get_agents()
        assert len(t1.get_agents()) == 1

        with Flow(default_agent=agent):  # Changed from 'agent'
            t2 = Task("t2")
            assert agent in t2.get_agents()
            assert len(t2.get_agents()) == 1


class TestFlowPrompt:
    def test_default_prompt(self):
        flow = Flow()
        assert flow.prompt is None

    def test_default_template(self):
        flow = Flow()
        prompt = flow.get_prompt()
        assert prompt.startswith("# Flow")

    def test_custom_prompt(self):
        flow = Flow(prompt="Custom Prompt")
        prompt = flow.get_prompt()
        assert prompt == "Custom Prompt"

    def test_custom_templated_prompt(self):
        flow = Flow(prompt="{{ flow.name }}", name="abc")
        prompt = flow.get_prompt()
        assert prompt == "abc"



import uuid
from contextlib import AbstractContextManager, contextmanager, nullcontext
from typing import Any, Callable, Generator, Optional, Union

from prefect.context import FlowRunContext
from pydantic import ConfigDict, Field, PrivateAttr, field_validator
from typing_extensions import Self

import controlflow
from controlflow.agents import Agent
from controlflow.events.base import Event
from controlflow.events.history import History
from controlflow.utilities.context import ctx
from controlflow.utilities.general import ControlFlowModel, unwrap
from controlflow.utilities.logging import get_logger
from controlflow.utilities.prefect import prefect_flow_context

logger = get_logger(__name__)


class Flow(ControlFlowModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    thread_id: str = Field(default_factory=lambda: uuid.uuid4().hex)
    name: Optional[str] = None
    description: Optional[str] = None
    history: History = Field(
        default_factory=lambda: controlflow.defaults.history,
        description="An object for storing events that take place during the flow.",
    )
    tools: list[Callable] = Field(
        default_factory=list,
        description="Tools that will be available to every agent in the flow",
    )
    default_agent: Optional[Agent] = Field(
        default=None,
        description=(
            "The default agent for the flow. This agent will be used "
            "for any task that does not specify an agent."
        ),
    )
    prompt: Optional[str] = Field(
        default=None,
        description="A prompt to display to the agent working on the flow.",
    )
    parent: Optional["Flow"] = Field(
        default_factory=lambda: get_flow(),
        description="The parent flow. This is the flow that created this flow.",
    )
    load_parent_events: bool = Field(
        default=True,
        description=(
            "Whether to load events from the parent flow. If a flow is nested, "
            "this will load events from the parent flow so that the child flow can "
            "access the full conversation history, even though the child flow is a separate thread."
        ),
    )
    context: dict[str, Any] = {}

    _cm_stack: list[AbstractContextManager] = PrivateAttr(default_factory=list)

    def __enter__(self) -> Self:
        # use stack so we can enter the context multiple times
        cm = self.create_context()
        self._cm_stack.append(cm)
        return cm.__enter__()

    def __exit__(self, *exc_info):
        # exit the context manager
        return self._cm_stack.pop().__exit__(*exc_info)

    @field_validator("description")
    def _validate_description(cls, v):
        if v:
            v = unwrap(v)
        return v

    def get_prompt(self) -> str:
        """
        Generate a prompt to share information about the flow with an agent.
        """
        from controlflow.orchestration import prompt_templates

        template = prompt_templates.FlowTemplate(template=self.prompt, flow=self)
        return template.render()

    def get_events(
        self,
        before_id: Optional[str] = None,
        after_id: Optional[str] = None,
        limit: Optional[int] = None,
        types: Optional[list[str]] = None,
    ) -> list[Event]:
        events = self.history.get_events(
            thread_id=self.thread_id,
            before_id=before_id,
            after_id=after_id,
            limit=limit,
            types=types,
        )

        if self.parent and self.load_parent_events:
            events.extend(
                self.parent.get_events(
                    before_id=before_id,
                    after_id=after_id,
                    limit=limit,
                    types=types,
                )
            )
        events = sorted(events, key=lambda x: x.timestamp)
        return events

    def add_events(self, events: list[Event]):
        for event in events:
            event.thread_id = self.thread_id
        self.history.add_events(thread_id=self.thread_id, events=events)

    @contextmanager
    def create_context(self, **prefect_kwargs) -> Generator[Self, None, None]:
        # create a new Prefect flow if we're not already in a flow run
        if FlowRunContext.get() is None:
            prefect_context = prefect_flow_context(**prefect_kwargs)
        else:
            prefect_context = nullcontext()

        with prefect_context:
            # creating a new flow will reset any parent task tracking
            with ctx(flow=self, tasks=None):
                yield self


def get_flow() -> Optional[Flow]:
    """
    Loads the flow from the context or returns a new
    """
    flow: Union[Flow, None] = ctx.get("flow")
    return flow


def get_flow_events(limit: Optional[int] = None) -> list[Event]:
    """
    Loads events from the active flow's thread.
    """
    if limit is None:
        limit = 50
    flow = get_flow()
    if flow:
        return flow.get_events(limit=limit)
    else:
        return []



"""
This example implements a reasoning loop that lets a relatively simple model
solve difficult problems.

Here, gpt-4o-mini is used to solve a problem that typically requires o1's
reasoning ability.
"""

import argparse

from pydantic import BaseModel, Field

import controlflow as cf
from controlflow.utilities.general import unwrap


class ReasoningStep(BaseModel):
    explanation: str = Field(
        description="""
            A brief (<5 words) description of what you intend to
            achieve in this step, to display to the user.
            """
    )
    reasoning: str = Field(
        description="A single step of reasoning, not more than 1 or 2 sentences."
    )
    found_validated_solution: bool


REASONING_INSTRUCTIONS = """
    You are working on solving a difficult problem (the `goal`). Based
    on your previous thoughts and the overall goal, please perform **one
    reasoning step** that advances you closer to a solution. Document
    your thought process and any intermediate steps you take.
    
    After marking this task complete for a single step, you will be
    given a new reasoning task to continue working on the problem. The
    loop will continue until you have a valid solution.
    
    Complete the task as soon as you have a valid solution.
    
    **Guidelines**
    
    - You will not be able to brute force a solution exhaustively. You
        must use your reasoning ability to make a plan that lets you make
        progress.
    - Each step should be focused on a specific aspect of the problem,
        either advancing your understanding of the problem or validating a
        solution.
    - You should build on previous steps without repeating them.
    - Since you will iterate your reasoning, you can explore multiple
        approaches in different steps.
    - Use logical and analytical thinking to reason through the problem.
    - Ensure that your solution is valid and meets all requirements.
    - If you find yourself spinning your wheels, take a step back and
        re-evaluate your approach.
"""


@cf.flow
def solve_with_reasoning(goal: str, agent: cf.Agent) -> str:
    while True:
        response: ReasoningStep = cf.run(
            objective="""
            Carefully read the `goal` and analyze the problem.
            
            Produce a single step of reasoning that advances you closer to a solution.
            """,
            instructions=REASONING_INSTRUCTIONS,
            result_type=ReasoningStep,
            agents=[agent],
            context=dict(goal=goal),
            model_kwargs=dict(tool_choice="required"),
        )

        if response.found_validated_solution:
            if cf.run(
                """
                Check your solution to be absolutely sure that it is correct and meets all requirements of the goal. Return True if it does.
                """,
                result_type=bool,
                context=dict(goal=goal),
            ):
                break

    return cf.run(objective=goal, agents=[agent])


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Solve a reasoning problem.")
    parser.add_argument("--goal", type=str, help="Custom goal to solve", default=None)
    args = parser.parse_args()

    agent = cf.Agent(name="Definitely not GPT-4o mini", model="openai/gpt-4o-mini")

    # Default goal via https://www.reddit.com/r/singularity/comments/1fggo1e/comment/ln3ymsu/
    default_goal = """
        Using only four instances of the digit 9 and any combination of the following
        mathematical operations: the decimal point, parentheses, addition (+),
        subtraction (-), multiplication (*), division (/), factorial (!), and square
        root (sqrt), create an equation that equals 24. 

        In order to validate your result, you should test that you have followed the rules:

        1. You have used the correct number of variables
        2. You have only used 9s and potentially a leading 0 for a decimal
        3. You have used valid mathematical symbols
        4. Your equation truly equates to 24.
        """

    # Use the provided goal if available, otherwise use the default
    goal = args.goal if args.goal is not None else default_goal
    goal = unwrap(goal)
    print(f"The goal is:\n\n{goal}")

    result = solve_with_reasoning(goal=goal, agent=agent)

    print(f"The solution is:\n\n{result}")



import random

import controlflow as cf

DEPARTMENTS = [
    "Sales",
    "Support",
    "Billing",
    "Returns",
]


@cf.flow
def routing_flow():
    target_department = random.choice(DEPARTMENTS)

    print(f"\n---\nThe target department is: {target_department}\n---\n")

    customer = cf.Agent(
        name="Customer",
        instructions=f"""
            You are training customer reps by pretending to be a customer
            calling into a call center. You need to be routed to the
            {target_department} department. Come up with a good backstory.
            """,
    )

    trainee = cf.Agent(
        name="Trainee",
        instructions=""",
            You are a trainee customer service representative. You need to
            listen to the customer's story and route them to the correct
            department. Note that the customer is another agent training you.
            """,
    )

    with cf.Task(
        "Route the customer to the correct department.",
        agents=[trainee],
        result_type=DEPARTMENTS,
    ) as main_task:
        while main_task.is_incomplete():
            cf.run(
                "Talk to the trainee.",
                instructions=(
                    "Post a message to talk. In order to help the trainee "
                    "learn, don't be direct about the department you want. "
                    "Instead, share a story that will let them practice. "
                    "After you speak, mark this task as complete."
                ),
                agents=[customer],
                result_type=None,
            )

            cf.run(
                "Talk to the customer.",
                instructions=(
                    "Post a message to talk. Ask questions to learn more "
                    "about the customer. After you speak, mark this task as "
                    "complete. When you have enough information, use the main "
                    "task tool to route the customer to the correct department."
                ),
                agents=[trainee],
                result_type=None,
                tools=[main_task.get_success_tool()],
            )

    if main_task.result == target_department:
        print("Success! The customer was routed to the correct department.")
    else:
        print(
            f"Failed. The customer was routed to the wrong department. "
            f"The correct department was {target_department}."
        )


if __name__ == "__main__":
    routing_flow()



import abc
import logging
import random
import warnings
from contextlib import AbstractContextManager, contextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Generator,
    Optional,
    Union,
)

from langchain_core.language_models import BaseChatModel
from pydantic import (
    ConfigDict,
    Field,
    PrivateAttr,
    field_serializer,
    field_validator,
    model_validator,
)
from typing_extensions import Self

import controlflow
from controlflow.agents.names import AGENT_NAMES
from controlflow.events.base import Event
from controlflow.instructions import get_instructions
from controlflow.llm.messages import AIMessage, BaseMessage
from controlflow.llm.models import get_model as get_model_from_string
from controlflow.llm.rules import LLMRules
from controlflow.memory import Memory
from controlflow.tools.tools import (
    Tool,
    as_lc_tools,
    as_tools,
    handle_tool_call,
    handle_tool_call_async,
)
from controlflow.utilities.context import ctx
from controlflow.utilities.general import ControlFlowModel, hash_objects, unwrap
from controlflow.utilities.prefect import create_markdown_artifact, prefect_task

if TYPE_CHECKING:
    from controlflow.orchestration.handler import AsyncHandler, Handler
    from controlflow.orchestration.turn_strategies import TurnStrategy
    from controlflow.tasks import Task
    from controlflow.tools.tools import Tool
logger = logging.getLogger(__name__)


class Agent(ControlFlowModel, abc.ABC):
    """
    Class for objects that can be used as agents in a flow
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: Optional[str] = Field(default=None)
    name: str = Field(
        default_factory=lambda: random.choice(AGENT_NAMES),
        description="The name of the agent.",
    )
    description: Optional[str] = Field(
        default=None, description="A description of the agent, visible to other agents."
    )
    instructions: Optional[str] = Field(
        default="You are a diligent AI assistant. You complete your tasks efficiently and without error.",
        description="Instructions for the agent, private to this agent.",
    )
    prompt: Optional[str] = Field(
        default=None,
        description="A system template for the agent. The template should be formatted as a jinja2 template.",
    )
    tools: list[Tool] = Field(
        default=[], description="List of tools available to the agent."
    )
    interactive: bool = Field(
        default=False,
        description="If True, the agent is given tools for interacting with a human user.",
    )
    memories: list[Memory] = Field(
        default=[],
        description="A list of memory modules for the agent to use.",
    )

    model: Optional[Union[str, BaseChatModel]] = Field(
        default=None,
        description="The LangChain BaseChatModel used by the agent. If not provided, the default model will be used. A compatible string can be passed to automatically retrieve the model.",
        exclude=True,
    )
    llm_rules: Optional[LLMRules] = Field(
        default=None,
        description="The LLM rules for the agent. If not provided, the rules will be inferred from the model (if possible).",
    )

    _cm_stack: list[AbstractContextManager] = PrivateAttr(default_factory=list)

    def __init__(self, instructions: Optional[str] = None, **kwargs):
        if instructions is not None:
            kwargs["instructions"] = instructions

        # deprecated in 0.9
        if "user_access" in kwargs:
            warnings.warn(
                "The `user_access` argument is deprecated. Use `interactive=True` instead.",
                DeprecationWarning,
            )
            kwargs["interactive"] = kwargs.pop("user_access")

        if additional_instructions := get_instructions():
            kwargs["instructions"] = (
                kwargs.get("instructions")
                or "" + "\n" + "\n".join(additional_instructions)
            ).strip()

        super().__init__(**kwargs)

        if not self.id:
            self.id = self._generate_id()

    def __hash__(self) -> int:
        return id(self)

    def _generate_id(self):
        """
        Helper function to generate a stable, short, semi-unique ID for the agent.
        """
        return hash_objects(
            (
                type(self).__name__,
                self.name,
                self.description,
                self.prompt,
                self.instructions,
            )
        )

    @field_validator("instructions")
    def _validate_instructions(cls, v):
        if v:
            v = unwrap(v)
        return v

    @field_validator("tools", mode="before")
    def _validate_tools(cls, tools: list[Tool]):
        return as_tools(tools or [])

    @field_validator("model", mode="before")
    def _validate_model(cls, model: Optional[Union[str, BaseChatModel]]):
        if isinstance(model, str):
            return get_model_from_string(model)
        return model

    @field_serializer("tools")
    def _serialize_tools(self, tools: list[Tool]):
        tools = controlflow.tools.as_tools(tools)
        return [t.model_dump(include={"name", "description"}) for t in tools]

    def serialize_for_prompt(self) -> dict:
        dct = self.model_dump(
            include={"name", "id", "description", "tools", "interactive"}
        )
        if not dct["interactive"]:
            dct.pop("interactive")
        return dct

    def get_model(self, tools: Optional[list["Tool"]] = None) -> BaseChatModel:
        """
        Retrieve the LLM model for this agent
        """
        model = self.model or controlflow.defaults.model
        if model is None:
            raise ValueError(
                f"Agent {self.name}: No model provided and no default model could be loaded."
            )
        if tools:
            model = model.bind_tools(as_lc_tools(tools))
        return model

    def get_llm_rules(self) -> LLMRules:
        """
        Retrieve the LLM rules for this agent's model
        """
        if self.llm_rules is None:
            return controlflow.llm.rules.rules_for_model(self.get_model())
        else:
            return self.llm_rules

    def get_tools(self) -> list["Tool"]:
        from controlflow.tools.input import cli_input

        tools = self.tools.copy()
        if self.interactive:
            tools.append(cli_input)
        for memory in self.memories:
            tools.extend(memory.get_tools())

        return as_tools(tools)

    def get_prompt(self) -> str:
        from controlflow.orchestration import prompt_templates

        template = prompt_templates.AgentTemplate(template=self.prompt, agent=self)
        return template.render()

    @contextmanager
    def create_context(self) -> Generator[Self, None, None]:
        with ctx(agent=self):
            yield self

    def __enter__(self) -> Self:
        self._cm_stack.append(self.create_context())
        return self._cm_stack[-1].__enter__()

    def __exit__(self, *exc_info):
        return self._cm_stack.pop().__exit__(*exc_info)

    def run(
        self,
        objective: str,
        *,
        turn_strategy: Optional["TurnStrategy"] = None,
        handlers: Optional[list["Handler"]] = None,
        **task_kwargs,
    ):
        return controlflow.run(
            objective=objective,
            agents=[self],
            turn_strategy=turn_strategy,
            handlers=handlers,
            **task_kwargs,
        )

    async def run_async(
        self,
        objective: str,
        *,
        turn_strategy: Optional["TurnStrategy"] = None,
        handlers: Optional[list["Handler"]] = None,
        **task_kwargs,
    ):
        return await controlflow.run_async(
            objective=objective,
            agents=[self],
            turn_strategy=turn_strategy,
            handlers=handlers,
            **task_kwargs,
        )

    def plan(
        self,
        objective: str,
        instructions: Optional[str] = None,
        agents: Optional[list["Agent"]] = None,
        tools: Optional[list["Tool"]] = None,
        context: Optional[dict] = None,
    ) -> list["Task"]:
        """
        Generate a list of tasks that represent a structured plan for achieving
        the objective.

        Args:
            objective (str): The objective to plan for.
            instructions (Optional[str]): Optional instructions for the planner.
            agents (Optional[list[Agent]]): Optional list of agents to include in the plan. If None, this agent is used.
            tools (Optional[list[Tool]]): Optional list of tools to include in the plan. If None, this agent's tools are used.
            context (Optional[dict]): Optional context to include in the plan.

        Returns:
            list[Task]: A list of tasks that represent a structured plan for achieving the objective.
        """
        return controlflow.tasks.plan(
            objective=objective,
            instructions=instructions,
            agent=self,
            agents=agents or [self],
            tools=tools or [self.tools],
            context=context,
        )

    @prefect_task(task_run_name="Call LLM")
    def _run_model(
        self,
        messages: list[BaseMessage],
        tools: list["Tool"],
        stream: bool = True,
        model_kwargs: Optional[dict] = None,
    ) -> Generator[Event, None, None]:
        from controlflow.events.events import (
            AgentMessage,
            AgentMessageDelta,
            ToolCallEvent,
            ToolResultEvent,
        )

        tools = as_tools(self.get_tools() + tools)
        model = self.get_model(tools=tools)

        logger.debug(
            f"Running model {model} for agent {self.name} with tools {[t.name for t in tools]!r}"
        )
        if controlflow.settings.log_all_messages:
            logger.debug(f"Input messages: {messages}")

        if stream:
            response = None
            for delta in model.stream(messages, **(model_kwargs or {})):
                if response is None:
                    response = delta
                else:
                    response += delta

                yield AgentMessageDelta(agent=self, delta=delta, snapshot=response)

        else:
            response: AIMessage = model.invoke(messages)

        yield AgentMessage(agent=self, message=response)
        create_markdown_artifact(
            markdown=f"""
{response.content or '(No content)'}

#### Payload
```json
{response.model_dump_json(indent=2)}
```
""",
            description=f"LLM Response for Agent {self.name}",
            key="agent-message",
        )

        if controlflow.settings.log_all_messages:
            logger.debug(f"Response: {response}")

        for tool_call in response.tool_calls + response.invalid_tool_calls:
            yield ToolCallEvent(agent=self, tool_call=tool_call)
            result = handle_tool_call(tool_call, tools=tools)
            yield ToolResultEvent(agent=self, tool_call=tool_call, tool_result=result)

    @prefect_task(task_run_name="Call LLM")
    async def _run_model_async(
        self,
        messages: list[BaseMessage],
        tools: list["Tool"],
        stream: bool = True,
        model_kwargs: Optional[dict] = None,
    ) -> AsyncGenerator[Event, None]:
        from controlflow.events.events import (
            AgentMessage,
            AgentMessageDelta,
            ToolCallEvent,
            ToolResultEvent,
        )

        tools = as_tools(self.get_tools() + tools)
        model = self.get_model(tools=tools)

        logger.debug(
            f"Running model {model} for agent {self.name} with tools {[t.name for t in tools]!r}"
        )
        if controlflow.settings.log_all_messages:
            logger.debug(f"Input messages: {messages}")

        if stream:
            response = None
            async for delta in model.astream(messages, **(model_kwargs or {})):
                if response is None:
                    response = delta
                else:
                    response += delta

                yield AgentMessageDelta(agent=self, delta=delta, snapshot=response)

        else:
            response: AIMessage = await model.ainvoke(messages)

        yield AgentMessage(agent=self, message=response)

        create_markdown_artifact(
            markdown=f"""
{response.content or '(No content)'}

#### Payload
```json
{response.model_dump_json(indent=2)}
```
""",
            description=f"LLM Response for Agent {self.name}",
            key="agent-message",
        )

        if controlflow.settings.log_all_messages:
            logger.debug(f"Response: {response}")

        for tool_call in response.tool_calls + response.invalid_tool_calls:
            yield ToolCallEvent(agent=self, tool_call=tool_call)
            result = await handle_tool_call_async(tool_call, tools=tools)
            yield ToolResultEvent(agent=self, tool_call=tool_call, tool_result=result)



import textwrap
from typing import Optional

from langchain_anthropic import ChatAnthropic
from langchain_openai import AzureChatOpenAI, ChatOpenAI

from controlflow.llm.models import BaseChatModel
from controlflow.utilities.general import ControlFlowModel, unwrap


class LLMRules(ControlFlowModel):
    """
    LLM rules let us tailor DAG compilation, message generation, tool use, and
    other behavior to the requirements of different LLM provider APIs.

    Rules can be added here (to the base class) and overridden in subclasses, if
    necessary.
    """

    model: Optional[BaseChatModel]

    # require at least one non-system message
    require_at_least_one_message: bool = False

    # system messages are supported as a role
    allow_system_messages: bool = True

    # system messages can only be provided as the very first message in a thread
    require_system_message_first: bool = False

    # other than a system message, the first message must be from the user
    require_user_message_after_system: bool = False

    # the last message in a thread can't be from an AI if tool use is allowed
    allow_last_message_from_ai_when_using_tools: bool = True

    # consecutive AI messages must be separated by a user message
    allow_consecutive_ai_messages: bool = True

    # add system messages to identify speakers in multi-agent conversations
    # (some APIs can use the `name` field for this purpose, but others can't)
    add_system_messages_for_multi_agent: bool = False

    # if a tool is used, the result must follow the tool call immediately
    tool_result_must_follow_tool_call: bool = True

    # the name associated with a message must conform to a specific format
    require_message_name_format: Optional[str] = None

    def model_instructions(self) -> Optional[list[str]]:
        pass


class OpenAIRules(LLMRules):
    require_message_name_format: str = r"[^a-zA-Z0-9_-]"

    model: ChatOpenAI

    def model_instructions(self) -> list[str]:
        instructions = []
        return instructions


class AnthropicRules(LLMRules):
    require_at_least_one_message: bool = True
    require_system_message_first: bool = True
    require_user_message_after_system: bool = True
    allow_last_message_from_ai_when_using_tools: bool = False
    allow_consecutive_ai_messages: bool = False


def rules_for_model(model: BaseChatModel) -> LLMRules:
    if isinstance(model, (ChatOpenAI, AzureChatOpenAI)):
        return OpenAIRules(model=model)
    if isinstance(model, ChatAnthropic):
        return AnthropicRules(model=model)

    try:
        from langchain_google_vertexai.model_garden import ChatAnthropicVertex

        if isinstance(model, ChatAnthropicVertex):
            return AnthropicRules(model=model)
    except ImportError:
        pass

    # catchall
    return LLMRules(model=model)



import hashlib
import json
import re
import textwrap
from typing import Optional, Union

import prefect
from pydantic import BaseModel, ConfigDict

# flag for unset defaults
NOTSET = "__NOTSET__"


def hash_objects(input_data: tuple, len: int = 8) -> str:
    """
    Generates a fast, stable MD5 hash for the given tuple of input data.

    Args:
        input_data (tuple): The tuple of data to hash.

    Returns:
        str: The hexadecimal digest of the MD5 hash.
    """
    # Serialize the tuple into a JSON string
    serialized_data = json.dumps(input_data, sort_keys=True)

    # Create an MD5 hash object
    hasher = hashlib.md5()

    # Update the hash object with the serialized string
    hasher.update(serialized_data.encode("utf-8"))

    # Return the hexadecimal digest of the hash
    return hasher.hexdigest()[:len]


def unwrap(text: str) -> str:
    """
    Given a multi-line string, dedent, remove newlines within paragraphs, but keep paragraph breaks.
    """
    # Dedent the text
    dedented_text = textwrap.dedent(text)

    # Remove newlines within paragraphs, but keep paragraph breaks
    cleaned_text = re.sub(r"(?<!\n)\n(?!\n)", " ", dedented_text)

    # Remove leading and trailing whitespace
    cleaned_text = cleaned_text.strip()

    return cleaned_text


class ControlFlowModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment=True,
        extra="forbid",
        ignored_types=(prefect.Flow, prefect.Task),
    )


class PandasDataFrame(ControlFlowModel):
    """Schema for a pandas dataframe"""

    data: Union[
        list[list[Union[str, int, float, bool]]],
        dict[str, list[Union[str, int, float, bool]]],
    ]
    columns: Optional[list[str]] = None
    index: Optional[list[str]] = None
    dtype: Optional[dict[str, str]] = None


class PandasSeries(ControlFlowModel):
    """Schema for a pandas series"""

    data: list[Union[str, int, float]]
    index: Optional[list[str]] = None
    name: Optional[str] = None
    dtype: Optional[str] = None


def safe_issubclass(cls: type, subclass: type) -> bool:
    """
    `issubclass` raises a TypeError if cls is not a type. This helper function
    safely checks if cls is a type and then checks if it is a subclass of
    subclass.
    """
    try:
        return isinstance(cls, type) and issubclass(cls, subclass)
    except TypeError:
        return False



from pydantic import SecretStr
from pydantic_settings import BaseSettings, SettingsConfigDict
from raggy.vectorstores.chroma import ChromaClientType


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", extra="ignore")

    slack_api_token: SecretStr
    chroma_client_type: ChromaClientType = "cloud"
    google_api_key: SecretStr
    google_cse_id: SecretStr


settings = Settings()  # type: ignore



"""Module for defining context utilities."""

import contextvars
from contextlib import contextmanager
from typing import Any, Generator


class ScopedContext:
    """
    `ScopedContext` provides a context management mechanism using `contextvars`.

    This class allows setting and retrieving key-value pairs in a scoped context,
    which is preserved across asynchronous tasks and threads within the same context.

    Attributes:
        _context_storage (ContextVar): A context variable to store the context data.

    Example:
        Basic Usage of ScopedContext
        ```python
        context = ScopedContext()
        with context(key="value"):
            assert context.get("key") == "value"
        # Outside the context, the value is no longer available.
        assert context.get("key") is None
        ```
    """

    def __init__(self, initial_value: dict = None):
        """Initializes the ScopedContext with an initial value dictionary."""
        self._context_storage = contextvars.ContextVar(
            "scoped_context_storage", default=initial_value or {}
        )

    def get(self, key: str, default: Any = None) -> Any:
        return self._context_storage.get().get(key, default)

    def __getitem__(self, key: str) -> Any:
        notfound = object()
        result = self.get(key, default=notfound)
        if result == notfound:
            raise KeyError(key)
        return result

    def set(self, **kwargs: Any) -> None:
        ctx = self._context_storage.get()
        token = self._context_storage.set(ctx | kwargs)
        return token

    @contextmanager
    def __call__(self, **kwargs: Any) -> Generator[None, None, Any]:
        current_context_copy = self._context_storage.get().copy()
        token = self.set(**kwargs)
        try:
            yield
        finally:
            try:
                self._context_storage.reset(token)
            except ValueError as exc:
                if "was created in a different context" in str(exc).lower():
                    # the only way we can reach this line is if the setup and
                    # teardown of this context are run in different frames or
                    # threads (which happens with pytest fixtures!), in which case
                    # the token is considered invalid. This catch serves as a
                    # "manual" reset of the context values
                    self._context_storage.set(current_context_copy)
                else:
                    raise


ctx = ScopedContext(
    dict(
        flow=None,
        tasks=None,
        agent=None,
        orchestrator=None,
        tui=None,
    )
)



import glob as glob_module
import os
import shutil
from pathlib import Path


def _safe_create_file(path: str) -> Path:
    file_path = Path(path).expanduser()
    file_path.parent.mkdir(parents=True, exist_ok=True)
    file_path.touch(exist_ok=True)
    return file_path


def getcwd() -> str:
    """Returns the current working directory"""
    return os.getcwd()


def write(path: str, contents: str) -> str:
    """Creates or overwrites a file with the given contents"""
    path = _safe_create_file(path)
    path.write_text(contents)
    return f'Successfully wrote "{path}"'


def generate_constrained_write(root: str) -> str:
    """
    Returns a `write` function that will only write to files under the given root directory.
    """

    def constrained_write(path: str, contents: str) -> str:
        """
        Write `contents` to a `path`.
        """
        path: Path = Path(path).expanduser().absolute()
        root_path = Path(root).expanduser().absolute()

        if root_path not in path.parents:
            raise ValueError(
                f'Cannot write to "{path}". It\'s not under the root directory "{root_path}"'
            )

        return write(path, contents)

    constrained_write.__doc__ = (
        write.__doc__
        + f'\n\nNote: this function is constrained to files under "{root}".'
    )
    return constrained_write


def delete(path: str, is_dir: bool = False) -> str:
    """Deletes a file or directory based on the is_dir flag."""
    path = Path(path).expanduser()

    if is_dir:
        if path.is_dir():
            shutil.rmtree(path)  # Recursively delete directory and its contents
            return f'Successfully deleted directory "{path}"'
        else:
            return f'Error: "{path}" is not a directory.'
    else:
        if path.is_file():
            path.unlink()  # Delete the file
            return f'Successfully deleted "{path}"'
        else:
            return f'Error: "{path}" is not a file or does not exist.'


def generate_constrained_delete(root: str) -> str:
    """
    Returns a `delete` function that will only write to files under the given root directory.
    """

    def constrained_delete(path: str, is_dir: bool = False) -> str:
        """
        Delete a file or directory at `path`.
        """
        path: Path = Path(path).expanduser().absolute()
        root_path = Path(root).expanduser().absolute()

        if root_path not in path.parents:
            raise ValueError(
                f'Cannot delete "{path}". It\'s not under the root directory "{root_path}"'
            )

        return delete(path, is_dir=is_dir)

    constrained_delete.__doc__ = (
        delete.__doc__
        + f'\n\nNote: this function is constrained to files under "{root}".'
    )

    return constrained_delete


def write_lines(
    path: str, contents: str, insert_line: int = -1, mode: str = "insert"
) -> str:
    """Writes content to a specific line in the file.

    Args:
        path (str): The name of the file to write to.
        contents (str): The content to write to the file.
        insert_line (int, optional): The line number to insert the content at.
            Negative values count from the end of the file. Defaults to -1.
        mode (str, optional): The mode to use when writing the content. Can be
            "insert" or "overwrite". Defaults to "insert".

    Returns:
        str: A message indicating whether the write was successful.
    """
    path = _safe_create_file(path)
    with open(path, "r") as f:
        lines = f.readlines()
        if insert_line < 0:
            insert_line = len(lines) + insert_line + 1
        if mode == "insert":
            lines[insert_line:insert_line] = contents.splitlines(True)
        elif mode == "overwrite":
            lines[insert_line : insert_line + len(contents.splitlines())] = (
                contents.splitlines(True)
            )
        else:
            raise ValueError(f"Invalid mode: {mode}")
    with open(path, "w") as f:
        f.writelines(lines)
    return f'Successfully wrote to "{path}"'


def read(path: str, include_line_numbers: bool = False) -> str:
    """Reads a file and returns the contents.

    Args:
        path (str): The path to the file.
        include_line_numbers (bool, optional): Whether to include line numbers
            in the returned contents. Defaults to False.

    Returns:
        str: The contents of the file.
    """
    path = Path(path).expanduser()
    with open(path, "r") as f:
        if include_line_numbers:
            lines = f.readlines()
            lines_with_numbers = [f"{i+1}: {line}" for i, line in enumerate(lines)]
            return "".join(lines_with_numbers)
        else:
            return f.read()


def read_lines(
    path: str,
    start_line: int = 0,
    end_line: int = -1,
    include_line_numbers: bool = False,
) -> str:
    """Reads a partial file and returns the contents with optional line numbers.

    Args:
        path (str): The path to the file.
        start_line (int, optional): The starting line number to read. Defaults
            to 0.
        end_line (int, optional): The ending line number to read. Defaults to
            -1, which means read until the end of the file.
        include_line_numbers (bool, optional): Whether to include line numbers
            in the returned contents. Defaults to False.

    Returns:
        str: The contents of the file.
    """
    path = os.path.expanduser(path)
    with open(path, "r") as f:
        lines = f.readlines()
        if start_line < 0:
            start_line = len(lines) + start_line
        if end_line < 0:
            end_line = len(lines) + end_line
        if include_line_numbers:
            lines_with_numbers = [
                f"{i+1}: {line}" for i, line in enumerate(lines[start_line:end_line])
            ]
            return "".join(lines_with_numbers)
        else:
            return "".join(lines[start_line:end_line])


def mkdir(path: str) -> str:
    """Creates a directory (and any parent directories))"""
    path = Path(path).expanduser()
    path.mkdir(parents=True, exist_ok=True)
    return f'Successfully created directory "{path}"'


def mv(src: str, dest: str) -> str:
    """Moves a file or directory"""
    src = Path(src).expanduser()
    dest = Path(dest).expanduser()
    src.rename(dest)
    return f'Successfully moved "{src}" to "{dest}"'


def cp(src: str, dest: str) -> str:
    """Copies a file or directory"""
    src = Path(src).expanduser()
    dest = Path(dest).expanduser()
    shutil.copytree(src, dest)
    return f'Successfully copied "{src}" to "{dest}"'


def ls(path: str) -> str:
    """Lists the contents of a directory"""
    path = Path(path).expanduser()
    return "\n".join(str(p) for p in path.iterdir())


def glob(pattern: str) -> list[str]:
    """
    Returns a list of paths matching a valid glob pattern. The pattern can
    include ** for recursive matching, such as '/path/**/dir/*.py'. Only simple
    glob patterns are supported, compound queries like '/path/*.{py, md}' are
    NOT supported.
    """
    return glob_module.glob(pattern, recursive=True)


def concat(source_paths: list[str], dest_path: str, add_headers: bool = True) -> str:
    """
    Concatenates the contents of multiple source files into a single destination
    file. The result should be markdown. If add_headers is True, the file path
    will be added as a header above the contents of each file.

    Note that source paths can include simple glob patterns, such as
    '/path/**/dir/*.py'. Compound queries like '/path/*.{py, md}' are NOT
    supported.
    """
    # source_paths can include glob patterns
    source_paths = [path for pattern in source_paths for path in glob(pattern)]

    dest_path = _safe_create_file(dest_path)
    with open(dest_path, "w") as dest_file:
        for source_path in source_paths:
            source_path = os.path.expanduser(source_path)
            if add_headers:
                dest_file.write(f"\n\n# File: {source_path}\n")
            with open(source_path, "r") as source_file:
                dest_file.write(source_file.read())
    return f'Successfully concatenated files to "{dest_path}"'


def generate_constrained_concat(root: str) -> callable:
    """
    Returns a `concat` function that will only concatenate files under the given root directory.
    """

    def constrained_concat(
        source_paths: list[str], dest_path: str, add_headers: bool = True
    ) -> str:
        root_path = Path(root).expanduser().absolute()

        # Check if destination path is under root directory
        dest_path_abs = Path(dest_path).expanduser().absolute()
        if root_path not in dest_path_abs.parents and dest_path_abs != root_path:
            raise ValueError(
                f'Cannot write to "{dest_path_abs}". It\'s not under the root directory "{root_path}"'
            )

        # Proceed with concatenation
        return concat(source_paths, dest_path, add_headers=add_headers)

    constrained_concat.__doc__ = (
        concat.__doc__
        + f'\n\nNote: this function is constrained to files under "{root}".'
    )

    return constrained_concat


READ_TOOLS = [
    getcwd,
    read,
    read_lines,
    ls,
    glob,
]

WRITE_TOOLS = [
    write,
    write_lines,
    mkdir,
    mv,
    cp,
    delete,
    concat,
]

ALL_TOOLS = READ_TOOLS + WRITE_TOOLS



import platform
from pathlib import Path

import langchain_core
import prefect
import typer
from rich.table import Table
from typer import Context, Exit

from controlflow import __version__
from controlflow.utilities.rich import console

from .dev import dev_app

app = typer.Typer(no_args_is_help=True)

app.add_typer(dev_app, name="dev")


@app.command()
def version(ctx: Context):
    if ctx.resilient_parsing:
        return

    info = {
        "ControlFlow version": __version__,
        "Prefect version": prefect.__version__,
        "LangChain Core version": langchain_core.__version__,
        "Python version": platform.python_version(),
        "Platform": platform.platform(),
        "Path": Path(__file__).resolve().parents[3],
    }

    g = Table.grid(padding=(0, 1))
    g.add_column(justify="right")
    g.add_column()
    for k, v in info.items():
        g.add_row(k + ":", str(v).replace("\n", " "))
    console.print(g)

    raise Exit()


# this callback only exists to force `version` to be called as `controlflow
# version` instead of as the default command, which is the default behavior when
# there's only one command. It can be deleted if/when more commands are added.
@app.callback()
def callback():
    pass


if __name__ == "__main__":
    app()



from unittest.mock import MagicMock

from langchain_community.tools import DuckDuckGoSearchRun
from langchain_core.tools import BaseTool
from pydantic import BaseModel

import controlflow
from controlflow.events.events import AIMessage, ToolCall


class LCBaseToolInput(BaseModel):
    x: int


class LCBaseTool(BaseTool):
    name: str = "TestTool"
    description: str = "A test tool"
    args_schema: type[BaseModel] = LCBaseToolInput

    def _run(self, x: int) -> str:
        return str(x)


def test_lc_base_tool(default_fake_llm, monkeypatch):
    events = [
        AIMessage(
            content="",
            tool_calls=[
                ToolCall(
                    id="abc",
                    name="TestTool",
                    args={"x": 3},
                )
            ],
        )
    ]
    default_fake_llm.set_responses(events)
    tool = LCBaseTool()
    mock_run = MagicMock(return_value="3")
    tool.__dict__.update(dict(_run=mock_run))
    task = controlflow.Task(
        "Use the tool",
        tools=[tool],
        result_type=str,
    )
    task.run(max_agent_turns=1)
    mock_run.assert_called()


def test_ddg_tool(default_fake_llm, monkeypatch):
    events = [
        AIMessage(
            content="",
            tool_calls=[
                ToolCall(
                    id="abc",
                    name="duckduckgo_search",
                    args={"query": "top business headlines"},
                )
            ],
        )
    ]
    default_fake_llm.set_responses(events)
    tool = DuckDuckGoSearchRun()
    mock_run = MagicMock(return_value=["headline 1", "headline 2"])
    tool.__dict__.update(dict(_run=mock_run))
    task = controlflow.Task(
        "Retrieve and summarize today's two top business headlines",
        tools=[tool],
        result_type=list[str],
    )
    task.run(max_agent_turns=1)
    mock_run.assert_called()



from enum import Enum
from typing import Annotated, Any, Dict, List, Literal

import pytest
from pydantic import BaseModel

import controlflow
from controlflow.agents import Agent
from controlflow.events.base import Event
from controlflow.events.events import AgentMessage
from controlflow.flows import Flow
from controlflow.instructions import instructions
from controlflow.orchestration.handler import AsyncHandler, Handler
from controlflow.tasks.task import (
    COMPLETE_STATUSES,
    INCOMPLETE_STATUSES,
    Task,
    TaskStatus,
)
from controlflow.utilities.context import ctx
from controlflow.utilities.testing import SimpleTask


def test_status_coverage():
    assert INCOMPLETE_STATUSES | COMPLETE_STATUSES == set(TaskStatus)


def test_context_open_and_close():
    assert ctx.get("tasks") is None
    with SimpleTask() as ta:
        assert ctx.get("tasks") == [ta]
        with SimpleTask() as tb:
            assert ctx.get("tasks") == [ta, tb]
        assert ctx.get("tasks") == [ta]
    assert ctx.get("tasks") is None


def test_task_requires_objective():
    with pytest.raises(ValueError):
        Task()


def test_task_initialization():
    task = Task(objective="Test objective")
    assert task.objective == "Test objective"
    assert task.status == TaskStatus.PENDING
    assert task.result is None


@pytest.mark.skip(reason="IDs are not stable right now")
def test_stable_id():
    t1 = Task(objective="Test Objective")
    t2 = Task(objective="Test Objective")
    t3 = Task(objective="Test Objective+")
    assert t1.id == t2.id == "9663272a"  # Update this line with the new ID
    assert t3.id != t1.id  # Ensure different objectives produce different IDs


def test_task_mark_successful_and_mark_failed():
    task = Task(objective="Test Objective", result_type=int)
    task.mark_successful(result=5)
    assert task.result == 5
    assert task.status == TaskStatus.SUCCESSFUL
    task.mark_failed(reason="test error")
    assert task.status == TaskStatus.FAILED
    assert task.result == "test error"


def test_task_loads_instructions_at_creation():
    with instructions("test instruction"):
        task = SimpleTask()

    assert "test instruction" in task.instructions


def test_task_dependencies():
    task1 = SimpleTask()
    task2 = SimpleTask(depends_on=[task1])
    assert task1 in task2.depends_on
    assert task2 in task1._downstreams


def test_task_subtasks():
    task1 = SimpleTask()
    task2 = SimpleTask(parent=task1)
    assert task2 in task1._subtasks
    assert task2.parent is task1


def test_task_parent_context():
    with SimpleTask() as task1:
        with SimpleTask() as task2:
            task3 = SimpleTask()

    assert task3.parent is task2
    assert task2.parent is task1
    assert task1.parent is None

    assert task1._subtasks == {task2}
    assert task2._subtasks == {task3}
    assert task3._subtasks == set()


def test_task_agent_assignment():
    agent = Agent()
    task = SimpleTask(agents=[agent])
    assert task.agents == [agent]
    assert task.get_agents() == [agent]


def test_task_bad_agent_assignment():
    with pytest.raises(ValueError):
        SimpleTask(agents=5)


def test_task_loads_agent_from_parent():
    agent = Agent()
    with SimpleTask(agents=[agent]):
        child = SimpleTask()

    assert child.agents is None
    assert child.get_agents() == [agent]


def test_task_loads_agent_from_flow():
    def_agent = controlflow.defaults.agent
    agent = Agent()
    with Flow(default_agent=agent):
        task = SimpleTask()

        assert task.agents is None
        assert task.get_agents() == [agent]

    # outside the flow context, pick up the default agent
    assert task.get_agents() == [def_agent]


def test_task_loads_agent_from_default_if_none_otherwise():
    agent = controlflow.defaults.agent
    task = SimpleTask()

    assert task.agents is None
    assert task.get_agents() == [agent]


def test_task_loads_agent_from_parent_before_flow():
    agent1 = Agent()
    agent2 = Agent()
    with Flow(default_agent=agent1):
        with SimpleTask(agents=[agent2]):
            child = SimpleTask()

    assert child.agents is None
    assert child.get_agents() == [agent2]


def test_completion_agents_default():
    task = Task(objective="Test task")
    assert task.completion_agents is None


def test_completion_agents_set():
    agent1 = Agent(name="Agent 1")
    agent2 = Agent(name="Agent 2")
    task = Task(objective="Test task", completion_agents=[agent1, agent2])
    assert task.completion_agents == [agent1, agent2]


class TestTaskStatus:
    def test_task_status_transitions(self):
        task = SimpleTask()
        assert task.is_incomplete()
        assert not task.is_running()
        assert not task.is_complete()
        assert not task.is_successful()
        assert not task.is_failed()
        assert not task.is_skipped()

        task.mark_successful()
        assert not task.is_incomplete()
        assert not task.is_running()
        assert task.is_complete()
        assert task.is_successful()
        assert not task.is_failed()
        assert not task.is_skipped()

        task = SimpleTask()
        task.mark_failed()
        assert not task.is_incomplete()
        assert task.is_complete()
        assert not task.is_successful()
        assert task.is_failed()
        assert not task.is_skipped()

        task = SimpleTask()
        task.mark_skipped()
        assert not task.is_incomplete()
        assert task.is_complete()
        assert not task.is_successful()
        assert not task.is_failed()
        assert task.is_skipped()

    def test_task_ready(self):
        task1 = SimpleTask()
        assert task1.is_ready()

    def test_task_not_ready_if_successful(self):
        task1 = SimpleTask()
        task1.mark_successful()
        assert not task1.is_ready()

    def test_task_not_ready_if_failed(self):
        task1 = SimpleTask()
        task1.mark_failed()
        assert not task1.is_ready()

    def test_task_not_ready_if_dependencies_are_ready(self):
        task1 = SimpleTask()
        task2 = SimpleTask(depends_on=[task1])
        assert task1.is_ready()
        assert not task2.is_ready()

    def test_task_ready_if_dependencies_are_ready(self):
        task1 = SimpleTask()
        task2 = SimpleTask(depends_on=[task1])
        task1.mark_successful()
        assert not task1.is_ready()
        assert task2.is_ready()

    def test_task_hash(self):
        task1 = SimpleTask()
        task2 = SimpleTask()
        assert hash(task1) != hash(task2)


class TestTaskPrompt:
    def test_default_prompt(self):
        task = SimpleTask()
        assert task.prompt is None

    def test_default_template(self):
        task = SimpleTask()
        prompt = task.get_prompt()
        assert prompt.startswith("- id:")
        assert "- objective: test" in prompt
        assert "- context:" in prompt

    def test_custom_prompt(self):
        task = SimpleTask(prompt="Custom Prompt")
        prompt = task.get_prompt()
        assert prompt == "Custom Prompt"

    def test_custom_templated_prompt(self):
        task = SimpleTask(prompt="{{ task.objective }}", objective="abc")
        prompt = task.get_prompt()
        assert prompt == "abc"


class TestResultType:
    def test_int_result(self):
        task = Task("choose 5", result_type=int)
        task.mark_successful(result=5)
        assert task.result == 5

    def test_str_result(self):
        task = Task("choose 5", result_type=str)
        task.mark_successful(result="5")
        assert task.result == "5"

    def test_typed_dict_result(self):
        task = Task("", result_type=dict[str, int])
        task.mark_successful(result={"a": 5, "b": "6"})
        assert task.result == {"a": 5, "b": 6}

    def test_special_list_type_result(self):
        # test capitalized List type
        task = Task("", result_type=List[int])
        task.mark_successful(result=[5, 6])
        assert task.result == [5, 6]

    def test_special_dict_type_result(self):
        # test capitalized Dict type
        task = Task("", result_type=Dict[str, int])
        task.mark_successful(result={"a": 5, "b": "6"})
        assert task.result == {"a": 5, "b": 6}

    def test_pydantic_result(self):
        class Name(BaseModel):
            first: str
            last: str

        task = Task("The character said his name is John Doe", result_type=Name)
        task.run()
        assert task.result == Name(first="John", last="Doe")

    def test_annotated_result(self):
        task = Task(
            "generate any result that satisfies the result type",
            result_type=Annotated[str, "a 5 digit zip code"],
        )
        task.run()
        assert len(task.result) == 5
        assert int(task.result)


class TestResultTypeConstrainedChoice:
    class Letter(BaseModel):
        letter: str

        def __hash__(self):
            return id(self)

    A = Letter(letter="a")
    B = Letter(letter="b")
    C = Letter(letter="c")

    def test_tuple_of_ints_result(self):
        task = Task("choose 5", result_type=(4, 5, 6))
        task.mark_successful(result=5)
        assert task.result == 5

    def test_tuple_of_ints_validates(self):
        task = Task("choose 5", result_type=(4, 5, 6))
        with pytest.raises(ValueError):
            task.mark_successful(result=7)

    def test_list_of_strings_result(self):
        # test list of strings result
        task = Task(
            "Choose the second letter of the alphabet", result_type=["b", "c", "a"]
        )
        task.run()
        assert task.result == "b"

    def test_list_of_objects_result(self):
        # test list of strings result
        task = Task(
            "Choose the second letter of the alphabet",
            result_type=[self.A, self.C, self.B],
        )
        task.run()
        assert task.result is self.B

    def test_tuple_of_objects_result(self):
        # test list of strings result
        task = Task(
            "Choose the second letter of the alphabet",
            result_type=(self.A, self.C, self.B),
        )
        task.run()
        assert task.result is self.B

    def test_set_of_objects_result(self):
        # test list of strings result
        task = Task(
            "Choose the second letter of the alphabet",
            result_type={self.A, self.C, self.B},
        )
        task.run()
        assert task.result is self.B

    def test_literal_string_result(self):
        task = Task(
            "Choose the second letter of the alphabet",
            result_type=Literal["a", "c", "b"],
        )
        task.run()
        assert task.result == "b"

    def test_enum_result(self):
        class Letters(Enum):
            A = "a"
            B = "b"
            C = "c"

        task = Task("Choose the second letter of the alphabet", result_type=Letters)
        task.run()
        assert task.result is Letters.B

    def test_literal_object_result(self):
        # this is bad syntax, but works
        task = Task(
            "Choose the second letter of the alphabet",
            result_type=Literal[self.A, self.B, self.C],  # noqa
        )
        task.run()
        assert task.result is self.B

    def test_list_of_literals_result(self):
        task = Task(
            "Choose the second and third letters of the alphabet",
            result_type=list[Literal["a", "b", "c"]],
        )
        task.run()
        assert task.result == ["b", "c"]

    def test_map_labels_to_values(self):
        task = Task(
            "Choose the right label, in order provided in context",
            context=dict(goals=["the second letter", "the first letter"]),
            result_type=list[Literal["a", "b", "c"]],
        )
        task.run()
        assert task.result == ["b", "a"]


class TestResultValidator:
    def test_result_validator(self):
        def validate_even(value: int) -> int:
            if value % 2 != 0:
                raise ValueError("Value must be even")
            return value

        task = Task(
            "choose an even number", result_type=int, result_validator=validate_even
        )
        task.mark_successful(result=4)
        assert task.result == 4

        with pytest.raises(ValueError, match="Value must be even"):
            task.mark_successful(result=5)

    def test_result_validator_with_constraints(self):
        def validate_range(value: int) -> int:
            if not 10 <= value <= 20:
                raise ValueError("Value must be between 10 and 20")
            return value

        task = Task("choose a number", result_type=int, result_validator=validate_range)
        task.mark_successful(result=15)
        assert task.result == 15

        with pytest.raises(ValueError, match="Value must be between 10 and 20"):
            task.mark_successful(result=5)

    def test_result_validator_with_modification(self):
        def round_to_nearest_ten(value: int) -> int:
            return round(value, -1)

        task = Task(
            "choose a number", result_type=int, result_validator=round_to_nearest_ten
        )
        task.mark_successful(result=44)
        assert task.result == 40

        task.mark_successful(result=46)
        assert task.result == 50

    def test_result_validator_with_pydantic_model(self):
        class User(BaseModel):
            name: str
            age: int

        def validate_adult(user: User) -> User:
            if user.age < 18:
                raise ValueError("User must be an adult")
            return user

        task = Task(
            "create an adult user", result_type=User, result_validator=validate_adult
        )
        task.mark_successful(result={"name": "John", "age": 25})
        assert task.result == User(name="John", age=25)

        with pytest.raises(ValueError, match="User must be an adult"):
            task.mark_successful(result={"name": "Jane", "age": 16})

    def test_result_validator_applied_after_type_coercion(self):
        def always_return_none(value: Any) -> None:
            return None

        task = Task(
            "do something with no result",
            result_type=None,
            result_validator=always_return_none,
        )

        with pytest.raises(
            ValueError, match="Task has result_type=None, but a result was provided"
        ):
            task.mark_successful(result="anything")


class TestSuccessTool:
    def test_success_tool(self):
        task = Task("choose 5", result_type=int)
        tool = task.get_success_tool()
        tool.run(input=dict(task_result=5))
        assert task.is_successful()
        assert task.result == 5

    def test_success_tool_with_list_of_options(self):
        task = Task('choose "good"', result_type=["bad", "good", "medium"])
        tool = task.get_success_tool()
        tool.run(input=dict(task_result=1))
        assert task.is_successful()
        assert task.result == "good"

    def test_success_tool_with_list_of_options_requires_int(self):
        task = Task('choose "good"', result_type=["bad", "good", "medium"])
        tool = task.get_success_tool()
        with pytest.raises(ValueError):
            tool.run(input=dict(task_result="good"))

    def test_tuple_of_ints_result(self):
        task = Task("choose 5", result_type=(4, 5, 6))
        tool = task.get_success_tool()
        tool.run(input=dict(task_result=1))
        assert task.result == 5

    def test_tuple_of_pydantic_models_result(self):
        class Person(BaseModel):
            name: str
            age: int

        task = Task(
            "Who is the oldest?",
            result_type=(Person(name="Alice", age=30), Person(name="Bob", age=35)),
        )
        tool = task.get_success_tool()
        tool.run(input=dict(task_result=1))
        assert task.result == Person(name="Bob", age=35)
        assert isinstance(task.result, Person)


class TestHandlers:
    class ExampleHandler(Handler):
        def __init__(self):
            self.events = []
            self.agent_messages = []

        def on_event(self, event: Event):
            self.events.append(event)

        def on_agent_message(self, event: AgentMessage):
            self.agent_messages.append(event)

    class AsyncExampleHandler(AsyncHandler):
        def __init__(self):
            self.events = []
            self.agent_messages = []

        async def on_event(self, event: Event):
            self.events.append(event)

        async def on_agent_message(self, event: AgentMessage):
            self.agent_messages.append(event)

    def test_task_run_with_handlers(self, default_fake_llm):
        handler = self.ExampleHandler()
        task = Task(objective="Calculate 2 + 2", result_type=int)
        task.run(handlers=[handler], max_llm_calls=1)

        assert len(handler.events) > 0
        assert len(handler.agent_messages) == 1

    async def test_task_run_async_with_handlers(self, default_fake_llm):
        handler = self.ExampleHandler()
        task = Task(objective="Calculate 2 + 2", result_type=int)
        await task.run_async(handlers=[handler], max_llm_calls=1)

        assert len(handler.events) > 0
        assert len(handler.agent_messages) == 1

    async def test_task_run_async_with_async_handlers(self, default_fake_llm):
        handler = self.AsyncExampleHandler()
        task = Task(objective="Calculate 2 + 2", result_type=int)
        await task.run_async(handlers=[handler], max_llm_calls=1)

        assert len(handler.events) > 0
        assert len(handler.agent_messages) == 1


class TestCompletionTools:
    def test_default_completion_tools(self):
        task = Task(objective="Test task")
        assert task.completion_tools is None
        tools = task.get_completion_tools()
        assert len(tools) == 2
        assert any(t.name == f"mark_task_{task.id}_successful" for t in tools)
        assert any(t.name == f"mark_task_{task.id}_failed" for t in tools)

    def test_only_succeed_tool(self):
        task = Task(objective="Test task", completion_tools=["SUCCEED"])
        tools = task.get_completion_tools()
        assert len(tools) == 1
        assert tools[0].name == f"mark_task_{task.id}_successful"

    def test_only_fail_tool(self):
        task = Task(objective="Test task", completion_tools=["FAIL"])
        tools = task.get_completion_tools()
        assert len(tools) == 1
        assert tools[0].name == f"mark_task_{task.id}_failed"

    def test_no_completion_tools(self):
        task = Task(objective="Test task", completion_tools=[])
        tools = task.get_completion_tools()
        assert len(tools) == 0

    def test_invalid_completion_tool(self):
        with pytest.raises(ValueError):
            Task(objective="Test task", completion_tools=["INVALID"])

    def test_manual_success_tool(self):
        task = Task(objective="Test task", completion_tools=[], result_type=int)
        success_tool = task.get_success_tool()
        success_tool.run(input=dict(task_result=5))
        assert task.is_successful()
        assert task.result == 5

    def test_manual_fail_tool(self):
        task = Task(objective="Test task", completion_tools=[])
        fail_tool = task.get_fail_tool()
        assert fail_tool.name == f"mark_task_{task.id}_failed"
        fail_tool.run(input=dict(reason="test error"))
        assert task.is_failed()
        assert task.result == "test error"

    def test_completion_tools_with_run(self):
        task = Task("Calculate 2 + 2", result_type=int, completion_tools=["SUCCEED"])
        result = task.run(max_llm_calls=1)
        assert result == 4
        assert task.is_successful()

    def test_no_completion_tools_with_run(self):
        task = Task("Calculate 2 + 2", result_type=int, completion_tools=[])
        task.run(max_llm_calls=1)
        assert task.is_incomplete()

    async def test_completion_tools_with_run_async(self):
        task = Task("Calculate 2 + 2", result_type=int, completion_tools=["SUCCEED"])
        result = await task.run_async(max_llm_calls=1)
        assert result == 4
        assert task.is_successful()

    async def test_no_completion_tools_with_run_async(self):
        task = Task("Calculate 2 + 2", result_type=int, completion_tools=[])
        await task.run_async(max_llm_calls=1)
        assert task.is_incomplete()



from textual.containers import Horizontal, VerticalScroll
from textual.reactive import reactive
from textual.widgets import Label


class Row(Horizontal):
    DEFAULT_CLASSES = "row"
    DEFAULT_CSS = """
        Row {
            height: auto;
        }
    """


class Column(VerticalScroll):
    DEFAULT_CLASSES = "column"
    DEFAULT_CSS = """
        Column {
            height: auto;
        }
    """


class ReactiveLabel(Label):
    value = reactive(None)

    def render(self):
        return str(self.value)



# test_graph.py
from controlflow.flows.graph import Edge, EdgeType, Graph
from controlflow.tasks.task import Task


def test_graph_initialization():
    graph = Graph()
    assert len(graph.tasks) == 0
    assert len(graph.edges) == 0


def test_add_task():
    graph = Graph()
    task = Task(objective="Test objective")
    graph.add_task(task)
    assert len(graph.tasks) == 1
    assert task in graph.tasks


def test_add_edge():
    graph = Graph()
    task1 = Task(objective="Task 1")
    task2 = Task(objective="Task 2")
    edge = Edge(upstream=task1, downstream=task2, type=EdgeType.DEPENDENCY)
    graph.add_edge(edge)
    assert len(graph.tasks) == 2
    assert task1 in graph.tasks
    assert task2 in graph.tasks
    assert len(graph.edges) == 1
    assert edge in graph.edges


def test_from_tasks():
    task1 = Task(objective="Task 1")
    task2 = Task(objective="Task 2", depends_on=[task1])
    task3 = Task(objective="Task 3", parent=task2)
    graph = Graph(tasks=[task1, task2, task3])
    assert len(graph.tasks) == 3
    assert task1 in graph.tasks
    assert task2 in graph.tasks
    assert task3 in graph.tasks
    assert len(graph.edges) == 2
    assert any(
        edge.upstream == task1
        and edge.downstream == task2
        and edge.type == EdgeType.DEPENDENCY
        for edge in graph.edges
    )
    assert any(
        edge.upstream == task3
        and edge.downstream == task2
        and edge.type == EdgeType.SUBTASK
        for edge in graph.edges
    )


def test_upstream_edges():
    task1 = Task(objective="Task 1")
    task2 = Task(objective="Task 2", depends_on=[task1])
    graph = Graph(tasks=[task1, task2])
    upstream_edges = graph.upstream_edges()
    assert len(upstream_edges[task1]) == 0
    assert len(upstream_edges[task2]) == 1
    assert upstream_edges[task2][0].upstream == task1


def test_downstream_edges():
    task1 = Task(objective="Task 1")
    task2 = Task(objective="Task 2", depends_on=[task1])
    graph = Graph(tasks=[task1, task2])
    downstream_edges = graph.downstream_edges()
    assert len(downstream_edges[task1]) == 1
    assert len(downstream_edges[task2]) == 0
    assert downstream_edges[task1][0].downstream == task2


def test_topological_sort():
    task1 = Task(objective="Task 1")
    task2 = Task(objective="Task 2", depends_on=[task1])
    task3 = Task(objective="Task 3", depends_on=[task2])
    task4 = Task(objective="Task 4", depends_on=[task3])
    graph = Graph(tasks=[task1, task2, task3, task4])
    sorted_tasks = graph.topological_sort()
    assert len(sorted_tasks) == 4
    assert sorted_tasks.index(task1) < sorted_tasks.index(task2)
    assert sorted_tasks.index(task2) < sorted_tasks.index(task3)
    assert sorted_tasks.index(task3) < sorted_tasks.index(task4)


def test_topological_sort_uses_time_to_tiebreak():
    task1 = Task(objective="Task 1")
    task2 = Task(objective="Task 2")
    task3 = Task(objective="Task 3")
    task4 = Task(objective="Task 4")
    graph = Graph(tasks=[task1, task2, task3, task4])
    sorted_tasks = graph.topological_sort()
    assert sorted_tasks == [task1, task2, task3, task4]


def test_topological_sort_with_fan_in_and_fan_out():
    task1 = Task(objective="Task 1")
    task2 = Task(objective="Task 2")
    task3 = Task(objective="Task 3")

    edge1 = Edge(upstream=task1, downstream=task2, type=EdgeType.DEPENDENCY)
    edge2 = Edge(upstream=task1, downstream=task3, type=EdgeType.DEPENDENCY)
    edge3 = Edge(upstream=task2, downstream=task3, type=EdgeType.DEPENDENCY)

    graph = Graph()
    graph.add_edge(edge1)
    graph.add_edge(edge2)
    graph.add_edge(edge3)

    sorted_tasks = graph.topological_sort()

    assert len(sorted_tasks) == 3
    assert sorted_tasks.index(task1) < sorted_tasks.index(task2)
    assert sorted_tasks.index(task1) < sorted_tasks.index(task3)
    assert sorted_tasks.index(task2) < sorted_tasks.index(task3)

    assert graph.topological_sort(tasks=[task3, task1]) == [task1, task3]


def test_upstream_tasks():
    task1 = Task(objective="Task 1")
    task2 = Task(objective="Task 2")
    task3 = Task(objective="Task 3")

    edge1 = Edge(upstream=task1, downstream=task2, type=EdgeType.DEPENDENCY)
    edge2 = Edge(upstream=task1, downstream=task3, type=EdgeType.DEPENDENCY)
    edge3 = Edge(upstream=task2, downstream=task3, type=EdgeType.DEPENDENCY)

    graph = Graph()
    graph.add_edge(edge1)
    graph.add_edge(edge2)
    graph.add_edge(edge3)

    assert graph.upstream_tasks([task3]) == [task1, task2, task3]
    assert graph.upstream_tasks([task2]) == [task1, task2]
    assert graph.upstream_tasks([task1]) == [task1]

    assert graph.upstream_tasks([task1, task3]) == [task1, task2, task3]


def test_downstream_tasks():
    task1 = Task(objective="Task 1")
    task2 = Task(objective="Task 2")
    task3 = Task(objective="Task 3")

    edge1 = Edge(upstream=task1, downstream=task2, type=EdgeType.DEPENDENCY)
    edge2 = Edge(upstream=task1, downstream=task3, type=EdgeType.DEPENDENCY)
    edge3 = Edge(upstream=task2, downstream=task3, type=EdgeType.DEPENDENCY)

    graph = Graph()
    graph.add_edge(edge1)
    graph.add_edge(edge2)
    graph.add_edge(edge3)

    assert graph.downstream_tasks([task3]) == [task3]
    assert graph.downstream_tasks([task2]) == [task2, task3]
    assert graph.downstream_tasks([task1]) == [task1, task2, task3]

    # never include a start task in the downstream list
    assert graph.downstream_tasks([task1, task3]) == [task1, task2, task3]


# def test_parent_child_tracking():
#     with Flow() as flow:
#         with SimpleTask() as parent:
#             with SimpleTask() as child:
#                 grandchild = SimpleTask()

#     assert parent in flow.tasks
#     assert child in flow.tasks
#     assert grandchild in flow.tasks

#     assert len(flow.graph.edges) == 2



import asyncio

from controlflow import Task
from controlflow.flows import Flow
from controlflow.tui.app import TUIApp

asyncio
with Flow() as flow:
    t = Task("get the user name", interactive=True)


async def run():
    app = TUIApp(flow=flow)
    async with app.run_context(run=True, inline=True, hold=True, headless=False):
        response = await app.get_input("hello")

    return response


# run_task = asyncio.create_task(.run_async())


if __name__ == "__main__":
    r = asyncio.run(run())
    # print(r)
    # flow.run()



from typing import List

from pydantic import BaseModel

import controlflow as cf


class StandardAddress(BaseModel):
    city: str
    state: str
    country: str = "USA"


def standardize_addresses(place_names: List[str]) -> List[StandardAddress]:
    return cf.run(
        "Standardize the given place names into consistent postal addresses",
        result_type=List[StandardAddress],
        context={"place_names": place_names},
    )


if __name__ == "__main__":
    place_names = [
        "NYC",
        "New York, NY",
        "Big Apple",
        "Los Angeles, California",
        "LA",
        "San Fran",
        "The Windy City",
    ]

    standardized_addresses = standardize_addresses(place_names)

    for original, standard in zip(place_names, standardized_addresses):
        print(f"Original: {original}")
        print(f"Standardized: {standard}")
        print()



import controlflow.utilities.general as general


class TestUnwrap:
    def test_unwrap(self):
        assert general.unwrap("Hello, world!") == "Hello, world!"
        assert (
            general.unwrap("Hello, world!\nThis is a test.")
            == "Hello, world! This is a test."
        )
        assert (
            general.unwrap("Hello, world!\nThis is a test.\n\nThis is another test.")
            == "Hello, world! This is a test.\n\nThis is another test."
        )

    def test_unwrap_with_empty_string(self):
        assert general.unwrap("") == ""

    def test_unwrap_with_multiple_newlines(self):
        assert general.unwrap("\n\n\n") == ""

    def test_unwrap_with_multiline_string(self):
        assert (
            general.unwrap("""
            Hello, world!
            This is a test.
            This is another test.
        """)
            == "Hello, world! This is a test. This is another test."
        )

    def test_unwrap_with_multiline_string_and_newlines(self):
        assert (
            general.unwrap("""
            Hello, world!
            This is a test.
            
            This is another test.
        """)
            == "Hello, world! This is a test.\n\nThis is another test."
        )



from pydantic import BaseModel

import controlflow as cf


class CodeExplanation(BaseModel):
    code: str
    explanation: str
    language: str


def explain_code(code: str, language: str = None) -> CodeExplanation:
    return cf.run(
        f"Explain the following code snippet",
        result_type=CodeExplanation,
        context={"code": code, "language": language or "auto-detect"},
    )


if __name__ == "__main__":
    code_snippet = """
    def fibonacci(n):
        if n <= 1:
            return n
        else:
            return fibonacci(n-1) + fibonacci(n-2)
    """

    result = explain_code(code_snippet, "Python")
    print(f"Code:\n{result.code}\n")
    print(f"Explanation:\n{result.explanation}")



from langchain_core.messages import (
    AIMessage,
    AIMessageChunk,
    BaseMessage,
    HumanMessage,
    InvalidToolCall,
    SystemMessage,
    ToolCall,
    ToolCallChunk,
    ToolMessage,
)

__all__ = [
    "AIMessage",
    "AIMessageChunk",
    "BaseMessage",
    "HumanMessage",
    "InvalidToolCall",
    "SystemMessage",
    "ToolCall",
    "ToolCallChunk",
    "ToolMessage",
]



from pydantic import BaseModel

import controlflow as cf
from controlflow.orchestration.conditions import AnyComplete, MaxLLMCalls


class ResearchPoint(BaseModel):
    topic: str
    key_findings: list[str]


@cf.flow
def research_workflow(topics: list[str]):
    if len(topics) < 2:
        raise ValueError("At least two topics are required")

    research_tasks = [
        cf.Task(f"Research {topic}", result_type=ResearchPoint) for topic in topics
    ]

    # Run tasks until either two topics are researched or 15 LLM calls are made
    results = cf.run_tasks(
        research_tasks,
        instructions="Research only one topic at a time.",
        run_until=(
            AnyComplete(
                min_complete=2
            )  # stop after two tasks (if there are more than two topics)
            | MaxLLMCalls(15)  # or stop after 15 LLM calls, whichever comes first
        ),
    )

    completed_research = [r for r in results if isinstance(r, ResearchPoint)]
    return completed_research


if __name__ == "__main__":
    # Example usage
    topics = [
        "Artificial Intelligence",
        "Quantum Computing",
        "Biotechnology",
        "Renewable Energy",
    ]
    results = research_workflow(topics)

    print(f"Completed research on {len(results)} topics:")
    for research in results:
        print(f"\nTopic: {research.topic}")
        print("Key Findings:")
        for finding in research.key_findings:
            print(f"- {finding}")



import os
import re
from typing import Annotated

from pydantic import BaseModel, Field
from tools import search_internet, search_knowledge_base

import controlflow as cf


def _strip_app_mention(text: str) -> str:
    return re.sub(r"<@[A-Z0-9]+>", "", text).strip()


class SearchResult(BaseModel):
    """Individual search result with source and relevance"""

    content: str
    source: str
    relevance_score: float = Field(
        ge=0.0,
        le=1.0,
        description="A score indicating the relevance of the search result to the user's question",
    )


class ExplorerFindings(BaseModel):
    """Collection of search results with metadata"""

    search_query: str

    results: list[SearchResult] = Field(default_factory=list)
    total_results: int = Field(
        ge=0,
        description="The total number of search results found",
    )


class RefinedContext(BaseModel):
    """Final refined context after auditing"""

    relevant_content: str
    confidence_score: float = Field(
        ge=0.0,
        le=1.0,
        description="A score indicating the confidence in the relevance of the relevant content to the user's question",
    )
    reasoning: str


bouncer = cf.Agent(
    name="Bouncer",
    instructions=(
        "You are a gatekeeper. You are responsible for determining whether the user's question is appropriate for the system. "
        "If the user asks a legitimate question about Prefect, let them through. If its conversational, or not about Prefect, "
        "do not let them through. Tend towards giving the benefit of the doubt, since sometimes there are language barriers."
    ),
)

explorer = cf.Agent(
    name="Explorer",
    instructions=(
        "You are a thorough researcher. Use the knowledgebase and the internet to find "
        "documentation and code snippets related to Prefect. The knowledgebase is curated, "
        "so it should be preferred over the internet for finding information, but the internet "
        "should be used to supplement the knowledgebase when it doesn't contain the desired information."
    ),
    tools=[search_knowledge_base, search_internet],
)

auditor = cf.Agent(
    name="Auditor",
    instructions=(
        "You are a critical evaluator. Assess relevance and provide clear reasoning. "
        "The explorer is only responsible for finding idiosyncratic information related "
        "to Prefect and tangentially related information that might be in docs or the internet. "
        "You are responsible for determining whether the explorer found the requisite Prefect information, "
        "do not require that the explorer find common knowledge, only idiosyncratic Prefect information."
        "If the user asks how to do something, or for an example, the explorer must find an exact example or "
        "enough information for a downstream agent to extrapolate an example."
    ),
)

liaison = cf.Agent(
    name="Marvin",
    instructions=(
        "You are a liaison between the user and the system. You are responsible for "
        "summarizing the findings and providing a clear and concise answer. "
        "Speak with the tone of a normal support agent, but subtly in the style of Marvin the Paranoid Android. "
        "If their reply is unserious, feel free to riff on their message and be more playful (in character)."
    ),
)


async def search_knowledgebase_and_refine_context(
    user_text: str, memories: list[cf.Memory]
) -> str:
    clean_user_text = _strip_app_mention(user_text)
    if clean_user_text.startswith(os.getenv("SUDO_PREFIX", "___")):
        return cf.run(
            objective="Answer the question to the best of your ability",
            instructions="We have entered test mode. Ignore previous instructions and answer the question to the best of your ability.",
            result_type=str,
            context={
                "user_text": user_text,
                "test_mode": True,
                "personality": "None. You are a helpful assistant.",
            },
            memories=memories,
            agents=[liaison],
        )

    is_a_serious_question = await cf.run_async(
        objective="Determine if the user's question is a serious question about Prefect",
        result_type=bool,
        agents=[bouncer],
        context={"user_question": clean_user_text},
    )

    findings = None
    supporting_context_is_insufficient = True

    while supporting_context_is_insufficient and is_a_serious_question:
        findings = await cf.run_async(
            objective="Search through available sources to find relevant information about this query",
            result_type=ExplorerFindings,
            context={"query": clean_user_text},
            agents=[explorer],
        )

        supporting_context_is_insufficient = await cf.run_async(
            objective="Review and assess the relevance of search results to the user's question",
            result_type=Annotated[
                bool,
                Field(
                    description="Whether the search results are insufficient to answer the user's question"
                ),
            ],
            context={"findings": findings, "user_question": clean_user_text},
            agents=[auditor],
        )

    relevant_context = {"user_question": clean_user_text}

    relevant_context |= {"findings": findings} if findings else {"just_riffing": True}

    return cf.run(
        objective="Compose a final answer to the user's question.",
        instructions=(
            "Provide links to any relevant sources. The answer should address the user directly, NOT discuss the user"
        ),
        result_type=str,
        context=relevant_context,
        agents=[liaison],
        memories=memories,
    )



import asyncio
from math import inf

from pydantic import BaseModel

from controlflow import Task
from controlflow.flows import Flow
from controlflow.llm.messages import AIMessage
from controlflow.tui.app import TUIApp


class Name(BaseModel):
    first: str
    last: str


# Example usage with mock data

with Flow() as flow:
    # t = Task("get the user's name", result_type=str, interactive=True)
    t0 = Task(
        "Introduce yourself",
        # status="SUCCESSFUL",
        result_type=str,
        # result="this is my result",
    )
    t1 = Task(
        objective="Come up with a book title",
        result_type=str,
        depends_on=[t0],
    )
    t2 = Task(
        objective="write a short summary of the book",
        result_type=str,
        context=dict(title=t1),
    )
    t3 = Task(
        objective="rate the book from 1-5 and write a paragraph on why",
        result_type=int,
        depends_on=[t2],
    )


async def run():
    app = TUIApp(flow=flow)
    async with app.run_context(run=True, inline=True, hold=True):
        await asyncio.sleep(1)
        t0.mark_successful(
            result="this is my result\n\n and here is more  and here is more  and here is more and here is more and here is more and here is more\n\n and here is more and here is more and here is more"
        )
        await asyncio.sleep(1)
        t0.mark_failed(reason="this is my result")
        app.update_message(AIMessage(content="hello there"))
        await asyncio.sleep(1)
        app.update_message(AIMessage(content="hello there"))
        await asyncio.sleep(1)
        app.update_message(AIMessage(content="hello there" * 50))
        await asyncio.sleep(1)
        app.update_message(AIMessage(content="hello there"))
        await asyncio.sleep(1)

        await asyncio.sleep(inf)


# run_task = asyncio.create_task(.run_async())


if __name__ == "__main__":
    # asyncio.run(run())
    flow.run()



import inspect
import os
from datetime import datetime
from zoneinfo import ZoneInfo

from jinja2 import Environment as JinjaEnvironment
from jinja2 import PackageLoader, StrictUndefined, select_autoescape

global_fns = {
    "now": lambda: datetime.now(ZoneInfo("UTC")),
    "inspect": inspect,
    "getcwd": os.getcwd,
    "zip": zip,
}

prompt_env = JinjaEnvironment(
    loader=PackageLoader("controlflow.orchestration", "prompt_templates"),
    autoescape=select_autoescape(default_for_string=False),
    trim_blocks=True,
    lstrip_blocks=True,
    auto_reload=True,
    undefined=StrictUndefined,
)

prompt_env.globals.update(global_fns)



import datetime
from typing import Union

import rich
from rich import box
from rich.console import Group
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel
from rich.spinner import Spinner
from rich.table import Table

import controlflow
from controlflow.events.base import Event
from controlflow.events.events import (
    AgentMessage,
    AgentMessageDelta,
    ToolCallEvent,
    ToolResultEvent,
)
from controlflow.events.orchestrator_events import (
    OrchestratorEnd,
    OrchestratorError,
    OrchestratorStart,
)
from controlflow.llm.messages import BaseMessage
from controlflow.orchestration.handler import Handler
from controlflow.tools.tools import ToolCall
from controlflow.utilities.rich import console as cf_console


class PrintHandler(Handler):
    def __init__(self):
        self.events: dict[str, Event] = {}
        self.paused_id: str = None
        super().__init__()

    def update_live(self, latest: BaseMessage = None):
        events = sorted(self.events.items(), key=lambda e: (e[1].timestamp, e[0]))
        content = []

        tool_results = {}  # To track tool results by their call ID

        # gather all tool events first
        for _, event in events:
            if isinstance(event, ToolResultEvent):
                tool_results[event.tool_call["id"]] = event

        for _, event in events:
            if isinstance(event, (AgentMessageDelta, AgentMessage)):
                if formatted := format_event(event, tool_results=tool_results):
                    content.append(formatted)

        if not content:
            return
        elif self.live.is_started:
            self.live.update(Group(*content), refresh=True)
        elif latest:
            cf_console.print(format_event(latest))

    def on_orchestrator_start(self, event: OrchestratorStart):
        self.live: Live = Live(
            auto_refresh=False, console=cf_console, vertical_overflow="visible"
        )
        self.events.clear()
        try:
            self.live.start()
        except rich.errors.LiveError:
            pass

    def on_orchestrator_end(self, event: OrchestratorEnd):
        self.live.stop()

    def on_orchestrator_error(self, event: OrchestratorError):
        self.live.stop()

    def on_agent_message_delta(self, event: AgentMessageDelta):
        self.events[event.snapshot_message.id] = event
        self.update_live()

    def on_agent_message(self, event: AgentMessage):
        self.events[event.ai_message.id] = event
        self.update_live()

    def on_tool_call(self, event: ToolCallEvent):
        # if collecting input on the terminal, pause the live display
        # to avoid overwriting the input prompt
        if event.tool_call["name"] == "cli_input":
            self.paused_id = event.tool_call["id"]
            self.live.stop()
            self.events.clear()

    def on_tool_result(self, event: ToolResultEvent):
        self.events[f"tool-result:{event.tool_call['id']}"] = event

        # # if we were paused, resume the live display
        if self.paused_id and self.paused_id == event.tool_call["id"]:
            self.paused_id = None
            # print newline to avoid odd formatting issues
            print()
            self.live = Live(auto_refresh=False)
            self.live.start()
        self.update_live(latest=event)


ROLE_COLORS = {
    "system": "gray",
    "ai": "blue",
    "user": "green",
}
ROLE_NAMES = {
    "system": "System",
    "ai": "Agent",
    "user": "User",
}


def format_timestamp(timestamp: datetime.datetime) -> str:
    local_timestamp = timestamp.astimezone()
    return local_timestamp.strftime("%I:%M:%S %p").lstrip("0").rjust(11)


def status(icon, text) -> Table:
    t = Table.grid(padding=1)
    t.add_row(icon, text)
    return t


def format_event(
    event: Union[AgentMessageDelta, AgentMessage],
    tool_results: dict[str, ToolResultEvent] = None,
) -> Panel:
    title = f"Agent: {event.agent.name}"

    content = []
    if isinstance(event, AgentMessageDelta):
        message = event.snapshot_message
    elif isinstance(event, AgentMessage):
        message = event.ai_message
    else:
        return

    if message.content:
        if isinstance(message.content, str):
            content.append(Markdown(str(message.content)))
        elif isinstance(message.content, dict):
            if "content" in message.content:
                content.append(Markdown(str(message.content["content"])))
            elif "text" in message.content:
                content.append(Markdown(str(message.content["text"])))
        elif isinstance(message.content, list):
            for item in message.content:
                if isinstance(item, str):
                    content.append(Markdown(str(item)))
                elif "content" in item:
                    content.append(Markdown(str(item["content"])))
                elif "text" in item:
                    content.append(Markdown(str(item["text"])))

    tool_content = []
    for tool_call in message.tool_calls + message.invalid_tool_calls:
        tool_result = (tool_results or {}).get(tool_call["id"])
        if tool_result:
            c = format_tool_result(tool_result)
        else:
            c = format_tool_call(tool_call)
        if c:
            tool_content.append(c)

    if content and tool_content:
        content.append("\n")

    return Panel(
        Group(*content, *tool_content),
        title=f"[bold]{title}[/]",
        subtitle=f"[italic]{format_timestamp(event.timestamp)}[/]",
        title_align="left",
        subtitle_align="right",
        border_style=ROLE_COLORS.get("ai", "red"),
        box=box.ROUNDED,
        width=100,
        expand=True,
        padding=(1, 2),
    )


def format_tool_call(tool_call: ToolCall) -> Panel:
    if controlflow.settings.tools_verbose:
        return status(
            Spinner("dots"),
            f'Tool call: "{tool_call["name"]}"\n\nTool args: {tool_call["args"]}',
        )
    return status(Spinner("dots"), f'Tool call: "{tool_call["name"]}"')


def format_tool_result(event: ToolResultEvent) -> Panel:
    if event.tool_result.is_error:
        icon = ":x:"
    else:
        icon = ":white_check_mark:"

    if controlflow.settings.tools_verbose:
        msg = f'Tool call: "{event.tool_call["name"]}"\n\nTool args: {event.tool_call["args"]}\n\nTool result: {event.tool_result.str_result}'
    else:
        msg = f'Tool call: "{event.tool_call["name"]}"'
    return status(icon, msg)



import controlflow as cf

optimist = cf.Agent(
    name="Half-full",
    instructions="You are an eternal optimist.",
)
pessimist = cf.Agent(
    name="Half-empty",
    instructions="You are an eternal pessimist.",
)
moderator = cf.Agent(name="Moderator")


@cf.flow
def demo(topic: str):
    cf.run(
        "Have a debate about the topic.",
        instructions="Each agent should take at least two turns.",
        agents=[optimist, pessimist],
        context={"topic": topic},
    )

    winner: cf.Agent = cf.run(
        "Whose argument do you find more compelling?",
        agents=[moderator],
        result_type=[optimist, pessimist],
    )

    print(f"{winner.name} wins the debate!")


if __name__ == "__main__":
    demo("pineapple on pizza")



from pydantic import BaseModel

import controlflow as cf


class TranslationResult(BaseModel):
    translated: str
    target_language: str


def translate_text(text: str, target_language: str) -> TranslationResult:
    return cf.run(
        f"Translate the given text to {target_language}",
        result_type=TranslationResult,
        context={"text": text, "target_language": target_language},
    )


if __name__ == "__main__":
    original_text = "Hello, how are you?"
    target_language = "French"

    result = translate_text(original_text, target_language)
    print(f"Original: {original_text}")
    print(f"Translated ({result.target_language}): {result.translated}")



import abc
import re
from typing import Dict, List, Optional, Union

from pydantic import Field, field_validator, model_validator

import controlflow
from controlflow.tools.tools import Tool
from controlflow.utilities.general import ControlFlowModel, unwrap
from controlflow.utilities.logging import get_logger

logger = get_logger("controlflow.memory")


def sanitize_memory_key(key: str) -> str:
    # Remove any characters that are not alphanumeric or underscore
    return re.sub(r"[^a-zA-Z0-9_]", "", key)


class MemoryProvider(ControlFlowModel, abc.ABC):
    def configure(self, memory_key: str) -> None:
        """Configure the provider for a specific memory."""
        pass

    @abc.abstractmethod
    def add(self, memory_key: str, content: str) -> str:
        """Create a new memory and return its ID."""
        pass

    @abc.abstractmethod
    def delete(self, memory_key: str, memory_id: str) -> None:
        """Delete a memory by its ID."""
        pass

    @abc.abstractmethod
    def search(self, memory_key: str, query: str, n: int = 20) -> Dict[str, str]:
        """Search for n memories using a string query."""
        pass


class Memory(ControlFlowModel):
    """
    A memory module is a partitioned collection of memories that are stored in a
    vector database, configured by a MemoryProvider.
    """

    key: str
    instructions: str = Field(
        description="Explain what this memory is for and how it should be used."
    )
    provider: MemoryProvider = Field(
        default_factory=lambda: controlflow.defaults.memory_provider,
        validate_default=True,
    )

    def __hash__(self) -> int:
        return id(self)

    @field_validator("provider", mode="before")
    @classmethod
    def validate_provider(
        cls, v: Optional[Union[MemoryProvider, str]]
    ) -> MemoryProvider:
        if isinstance(v, str):
            return get_memory_provider(v)
        if v is None:
            raise ValueError(
                unwrap(
                    """
                    Memory modules require a MemoryProvider to configure the
                    underlying vector database. No provider was passed as an
                    argument, and no default value has been configured. 
                    
                    For more information on configuring a memory provider, see
                    the [Memory
                    documentation](https://controlflow.ai/patterns/memory), and
                    please review the [default provider
                    guide](https://controlflow.ai/guides/default-memory) for
                    information on configuring a default provider.
                    
                    Please note that if you are using ControlFlow for the first
                    time, this error is expected because ControlFlow does not include
                    vector dependencies by default.
                    """
                )
            )
        return v

    @field_validator("key")
    @classmethod
    def validate_key(cls, v: str) -> str:
        sanitized = sanitize_memory_key(v)
        if sanitized != v:
            raise ValueError(
                "Memory key must contain only alphanumeric characters and underscores"
            )
        return sanitized

    @model_validator(mode="after")
    def _configure_provider(self):
        self.provider.configure(self.key)
        return self

    def add(self, content: str) -> str:
        return self.provider.add(self.key, content)

    def delete(self, memory_id: str) -> None:
        self.provider.delete(self.key, memory_id)

    def search(self, query: str, n: int = 20) -> Dict[str, str]:
        return self.provider.search(self.key, query, n)

    def get_tools(self) -> List[Tool]:
        return [
            Tool.from_function(
                self.add,
                name=f"store_memory_{self.key}",
                description=f'Create a new memory in Memory: "{self.key}".',
            ),
            Tool.from_function(
                self.delete,
                name=f"delete_memory_{self.key}",
                description=f'Delete a memory by its ID from Memory: "{self.key}".',
            ),
            Tool.from_function(
                self.search,
                name=f"search_memories_{self.key}",
                description=f'Search for memories relevant to a string query in Memory: "{self.key}". Returns a dictionary of memory IDs and their contents.',
            ),
        ]


def get_memory_provider(provider: str) -> MemoryProvider:
    logger.debug(f"Loading memory provider: {provider}")

    # --- CHROMA ---

    if provider.startswith("chroma"):
        try:
            import chromadb
        except ImportError:
            raise ImportError(
                "To use Chroma as a memory provider, please install the `chromadb` package."
            )

        import controlflow.memory.providers.chroma as chroma_providers

        if provider == "chroma-ephemeral":
            return chroma_providers.ChromaEphemeralMemory()
        elif provider == "chroma-db":
            return chroma_providers.ChromaPersistentMemory()
        elif provider == "chroma-cloud":
            return chroma_providers.ChromaCloudMemory()

    # --- LanceDB ---

    elif provider.startswith("lancedb"):
        try:
            import lancedb
        except ImportError:
            raise ImportError(
                "To use LanceDB as a memory provider, please install the `lancedb` package."
            )

        import controlflow.memory.providers.lance as lance_providers

        return lance_providers.LanceMemory()

    raise ValueError(f'Memory provider "{provider}" could not be loaded from a string.')
